! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2013 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!               Space-filling curve partitioning               !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Space-filling curve partitioning of an $fd4_domain.
!! Supports Hilbert and Morton SFC.
!! One-dimensional partitioning algorithms from $fd4_part_1d_mod are used.
!!
!! @see hilbert_mod
!! @see fd4_part_1d_mod
!! @see fd4_domain_mod
!! @author Matthias Lieber
module fd4_part_sfc_mod

  use fd4_globaldef_mod
  use fd4_domain_mod
  use fd4_iter_mod
  use fd4_part_1d_mod
  use fd4_mpi_mod
  use hilbert_mod
  use hilbert2d_mod
  use morton_mod
  use morton2d_mod
  use mrgrnk_mod
  use stack_mod
  use stack3_mod
  use rbtree_fd4_block_mod
  use timing_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
  
  integer(i_k), parameter, public :: FD4_PART_SFC_SIMPLE = 1
  integer(i_k), parameter, public :: FD4_PART_SFC_BINSRCH = 2
  integer(i_k), parameter, public :: FD4_PART_SFC_PARALLEL = 4
  integer(i_k), parameter, public :: FD4_PART_SFC_SIMPLE2 = 8
  integer(i_k), parameter, public :: FD4_PART_SFC_RB = 16
  integer(i_k), parameter, public :: FD4_PART_SFC_EBS = 32
  
  integer(i_k), parameter, public :: FD4_PART_SFC_HILBERT = 128
  integer(i_k), parameter, public :: FD4_PART_SFC_MORTON  = 256
  
  ! public methods
  public :: fd4_part_sfc_noweight
  public :: fd4_part_sfc_weight
  public :: fd4_part_sfc_weight2
  public :: fd4_part_sfc_update_blkmap_loc
  public :: fd4_part_sfc_update_blkmap
  public :: fd4_part_sfc_update_blkmap_bnd
  public :: fd4_part_sfc_prefixsum_parallel

  ! public types
  public :: part_sfc_iter
  
  !! general iterator over SFC
  type part_sfc_iter
    integer(i_k) :: p(3) = -1    ! spatial position
    integer(i_k) :: idx = -1     ! position on curve
    integer(i_k) :: sfctype = -1
    integer(i_k) :: map2d(2) = 1
    type(hilbert_iter) :: hiter
    type(hilbert2d_iter) :: h2diter
    type(morton_iter) :: miter
    type(morton2d_iter) :: m2diter
  end type part_sfc_iter
  
contains  

  include 'framework/fd4_domain.incf'
  include 'util/stack.incf'
  include 'util/stack3.incf'
  include 'util/math.incf'
  include 'framework/fd4_part_sfc.incf'


  !! Compute prefix sum of the current block weights (block%weight) in parallel.
  !!
  !! Stores result distributed in domain%balance%sfcd_blkweights_ps.
  !!
  !! The local sfcd_blkweights_ps(0) contains always the sum weight of all blocks
  !! on ranks < current rank.
  !!
  subroutine fd4_part_sfc_prefixsum_parallel(domain, sfctype, err, opt_sumweight, opt_maxweight, opt_consistent)
    type(fd4_domain), intent(inout) :: domain                         !! the domain
    integer(i_k)    , intent(in)    :: sfctype                        !! space-filling curve to use
    integer(i_k)    , intent(out)   :: err                            !! error status: 0...ok
    real(r8k)       , intent(out)   :: opt_sumweight                  !! sum of all block weights (bcast'ed to all)
    real(r8k)       , intent(out)   :: opt_maxweight                  !! local maximum weight
    logical         , intent(in)    :: opt_consistent                 !! ensure consistent values at process borders (default: no)
    target :: domain
    optional :: opt_sumweight, opt_maxweight, opt_consistent

    type(fd4_mpi), pointer :: mpi
    integer :: rank, nprocs, nblocks
    type(fd4_domain_balance), pointer :: bal
    real(r8k) :: localweight, localweight_ps
    type(fd4_iter) :: iter
    integer(i_k) :: ib
    integer(i8k) :: timing_us0mpi, timing_us1
    logical :: consistent

    err = -1

    mpi    => domain%mpi
    bal    => domain%balance
    rank    = mpi%rank
    nprocs  = mpi%nprocs
    nblocks = PRODUCT(domain%blnum(1:3))
    
    consistent = .false.
    if(present(opt_consistent)) consistent = opt_consistent

    ! allocate array for local block weights (twice the required size just as buffer for next time)
    if(.not. associated(bal%sfcd_blkweights_ps)) then
      allocate(bal%sfcd_blkweights_ps(0:2*domain%blocks%length), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate bal%sfcd_blkweights_ps')
      err = -1
    else
      if( ubound(bal%sfcd_blkweights_ps,1) <      domain%blocks%length  .or. &
          ubound(bal%sfcd_blkweights_ps,1) > 16 * domain%blocks%length ) then
        deallocate(bal%sfcd_blkweights_ps)
        allocate(bal%sfcd_blkweights_ps(0:2*domain%blocks%length), stat=err)
        FD4_L1_ASSERT(err/=0,'allocate bal%sfcd_blkweights_ps')
        err = -1
      end if
    end if

    ! get the local block weights and write them to bal%sfcd_blkweights_ps (ordered by SFC)
    ! and get the sum of all local block weights
    localweight = 0.0_r8k
    iter%cur => rbtree_fd4_block_iter_first(iter%biter, domain%blocks)
    if(present(opt_maxweight)) then
      opt_maxweight = 0.0_r8k
      do while(associated(iter%cur))
        call fd4_part_sfc_c2i_qsz(domain, sfctype, iter%cur%pos, ib)
        bal%sfcd_blkweights_ps(ib-bal%sfc_pstart+1) = REAL(iter%cur%weight, r8k)
        localweight = localweight + bal%sfcd_blkweights_ps(ib-bal%sfc_pstart+1)
        opt_maxweight = MAX(opt_maxweight, bal%sfcd_blkweights_ps(ib-bal%sfc_pstart+1))
        iter%cur => rbtree_fd4_block_iter_next(iter%biter)
      end do
    else
      do while(associated(iter%cur))
        call fd4_part_sfc_c2i_qsz(domain, sfctype, iter%cur%pos, ib)
        bal%sfcd_blkweights_ps(ib-bal%sfc_pstart+1) = REAL(iter%cur%weight, r8k)
        localweight = localweight + bal%sfcd_blkweights_ps(ib-bal%sfc_pstart+1)
        iter%cur => rbtree_fd4_block_iter_next(iter%biter)
      end do
    end if

    if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)

    ! parallel (exclusive) prefix sum of local block weight sums
    call MPI_Exscan(localweight, localweight_ps, 1, MPI_R8K, MPI_SUM, mpi%comm, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Exscan failed')
    err = -1

    ! start sending my prefix sum result to rank-1 to guarantee consistent values
    if(consistent .and. rank>0) then
      call MPI_Bsend(localweight_ps, 1, MPI_R8K, rank-1, 5, mpi%comm, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Bsend failed')
      err = -1
    end if

    if(domain%stat_enabled) then
      call gettime_ofday(timing_us1)
      domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
    end if

    ! rank 0 MPI_Exscan result is not defined
    if(rank==0) localweight_ps = 0.0_r8k

    ! apply (global) prefix sum to (local) weight array, i.e. start with MPI_Exscan result
    bal%sfcd_blkweights_ps(0) = localweight_ps
    do ib=1,domain%blocks%length
      bal%sfcd_blkweights_ps(ib) = bal%sfcd_blkweights_ps(ib) + bal%sfcd_blkweights_ps(ib-1)
    end do

    if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)
    
    ! receive the prefix sum result from rank+1 to set my last value in bal%sfcd_blkweights_ps
    if(consistent .and. rank<nprocs-1) then
      call MPI_Recv(bal%sfcd_blkweights_ps(domain%blocks%length), 1, MPI_R8K, rank+1, 5, mpi%comm, MPI_STATUS_IGNORE, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Recv failed')
      err = -1
    end if

    if(present(opt_sumweight)) then
      ! get sumweight from last global rank
      opt_sumweight = bal%sfcd_blkweights_ps(domain%blocks%length)
      call MPI_Bcast(opt_sumweight, 1, MPI_R8K, nprocs-1, mpi%comm, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Bcast failed')
      err = -1
    end if

    if(domain%stat_enabled) then
      call gettime_ofday(timing_us1)
      domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
    end if

    err = 0

  end subroutine fd4_part_sfc_prefixsum_parallel




  !! Create a new partition using block position. Considers block weights from domain%balance%blkweights.
  !!
  !! This version only works in non-adaptive block mode and when the SFC type is never changed.
  !!
  !! $balance is not set properly by all methods.
  !!
  !! Optimizations in comparison to $fd4_part_sfc_weight:
  !!  - sorting of the weight arrays is not necessary
  !!  - $domain%$balance%$blkweights is not required
  !!  - $domain%$balance%$rnk and $domain%$balance%$hidx are not required
  !!  - prefix sum can be computed in parallel with $par_pfxs=.true.
  !!
  !! Comments on $p1d_root:
  !!  - $p1d_root<0 means: all processes compute 1d partitioning
  !!  - $p1d_root>=0 implies $opt_par_pfxs=.true.
  !!  - $p1d_root>=0 does not work with method==FD4_PART_SFC_PARALLEL
  !!  - domain%balance%localweight is not set (set to 0.0) when using $p1d_root>=0
  !!
  !! $maxwgt_comm is only relevant for methods FD4_PART_SFC_BINSRCH, FD4_PART_SFC_PARALLEL, and FD4_PART_SFC_EBS.
  !!
  subroutine fd4_part_sfc_weight2(domain, local_blocks, method, sfctype, &
                                  binsteps, accuracy, blkmap_update, par_pfxs, p1d_root, maxwgt_comm, balance, err)
    type(fd4_domain), intent(inout) :: domain                  !! the domain
    type(stack3)    , intent(inout) :: local_blocks            !! stack3 of new local partition (differs from rank to rank)
    integer(i_k)    , intent(in)    :: method                  !! 1D partitioning method to use
    integer(i_k)    , intent(in)    :: sfctype                 !! space-filling curve to use
    integer(i_k)    , intent(in) :: binsteps                   !! max. number of steps for binary search methods
    !! target ratio from best possible load balance for binary steps estimation, *must be <= 1.0*
    real(r8k)       , intent(in) :: accuracy
    logical         , intent(in) :: blkmap_update              !! update the full block map
    logical         , intent(in) :: par_pfxs                   !! perform prefix sum in parallel
    integer(i_k)    , intent(in) :: p1d_root                   !! perform 1d-part. calc. serial on rank $p1d_root if $p1d_root>=0
    logical         , intent(in) :: maxwgt_comm                !! determine maxweight with MPI reduce (or local comp.)
    real(r4k)       , intent(out):: balance                    !! estimated balance of new partitioning
    integer(i_k)    , intent(out):: err                        !! error status: 0...ok
    target :: domain

    integer :: rank, nprocs, nblocks
    integer(i_k) :: ib, pos(3), oldp(2), pstart, pend, reqsteps, root
    real(r8k) :: maxweight, maxload, eps
    real(r8k), target :: dummy(0:1)
    real(r8k), pointer :: blkweights_ps(:) => NULL()   ! prefix sum of sfc sorted block weights
    integer(i_k), pointer :: partition(:) => NULL()    ! partition vector (start indices)
    type(fd4_iter) :: iter
    type(fd4_mpi), pointer :: mpi
    type(fd4_domain_balance), pointer :: bal
    type(part_sfc_iter) :: sfciter
    integer(i8k) :: mpi_time, timing_us0, timing_us0mpi, timing_us1
    logical :: do_par_pfxs
    integer :: get_maxweight
    
    balance = 0.0_r4k

    err = -1
    
    FD4_L1_ASSERT(domain%varinfo%vthres_mode,'not allowed in adaptive block mode')
    if(domain%blockcount>0 .and. sfctype/=domain%balance%last_sfctype) then
      FD4_L1_ERROR('cannot change SFC type')
    end if
    if(domain%blocks%length>0 .and. domain%balance%sfc_pstart<0) then
      FD4_L1_ERROR('domain%balance%sfc_pstart not set')
    end if

    mpi    => domain%mpi
    bal    => domain%balance
    rank    = mpi%rank
    nprocs  = mpi%nprocs
    nblocks = PRODUCT(domain%blnum(1:3))
    
    root = MIN(p1d_root, mpi%nprocs-1)
    do_par_pfxs = par_pfxs
    if(root>=0) do_par_pfxs = .true.
    
    if( method==FD4_PART_SFC_BINSRCH .or. method==FD4_PART_SFC_PARALLEL .or. method==FD4_PART_SFC_EBS ) then
      if(maxwgt_comm) then
        get_maxweight = 1
      else
        get_maxweight = 2
      end if
    else
      get_maxweight = 0
    end if
    
    if(method==FD4_PART_SFC_PARALLEL .and. root>=0) then
      FD4_L1_ERROR('FD4_PART_SFC_PARALLEL must be computed parallel')
    end if
    
    if(root<0 .or. rank==root) then  ! if root >= 0: only root allocates blkweights_ps
      if(.not. associated(bal%blkweights_ps)) then
        allocate(bal%blkweights_ps(0:PRODUCT(domain%blnum(1:3))), stat=err)
        FD4_L1_ASSERT(err/=0,'allocate bal%blkweights_ps')
      else
        if( ubound(bal%blkweights_ps,1) < PRODUCT(domain%blnum(1:3)) ) then
          deallocate(bal%blkweights_ps)
          allocate(bal%blkweights_ps(0:PRODUCT(domain%blnum(1:3))), stat=err)
          FD4_L1_ASSERT(err/=0,'allocate bal%blkweights_ps')
        end if
      end if
      blkweights_ps => bal%blkweights_ps
    end if
    
    if(.not. associated(bal%partition)) then
      allocate(bal%partition(0:domain%mpi%nprocs), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate bal%partition')
    end if
    partition => bal%partition

    if(domain%blockcount == nblocks) then
      !! all blocks already allocated
      
      if(do_par_pfxs) then
        !! parallel prefix sum of block weights + (all)gather of weights
      
        if(domain%stat_enabled) call gettime_ofday(timing_us0)
        
        ! compute parallel prefix sum of block weights
        ! stores result in bal%sfcd_blkweights_ps
        if(get_maxweight==1) then
          call fd4_part_sfc_prefixsum_parallel(domain, sfctype, err, opt_maxweight = maxweight)
        else
          call fd4_part_sfc_prefixsum_parallel(domain, sfctype, err)
        end if
        FD4_L1_ASSERT(err/=0,'fd4_part_sfc_prefixsum_parallel failed')
        
        if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)
        
        ! get global maxweight
        if(get_maxweight==1) then
          if(root<0) then
#ifdef FD4_NO_MPI_IN_PLACE
#warning compiling with FD4_NO_MPI_IN_PLACE
            dummy(0) = maxweight
            call MPI_Allreduce(dummy(0), maxweight, 1, MPI_R8K, MPI_MAX, mpi%comm, err)
#else
            call MPI_Allreduce(MPI_IN_PLACE, maxweight, 1, MPI_R8K, MPI_MAX, mpi%comm, err)
#endif
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Allreduce failed')
          else
            dummy(0) = maxweight
            call MPI_Reduce(dummy(0), maxweight, 1, MPI_R8K, MPI_MAX, root, mpi%comm, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Reduce failed')
          end if
        end if

        if(domain%stat_enabled) then
          call gettime_ofday(timing_us1)
          domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
          domain%stat%values(bal%stat_pfxs_us) = domain%stat%values(bal%stat_pfxs_us) + (timing_us1-timing_us0)
        end if

        ! collect the (prefix sum of) block weights to blkweights_ps (r8k array)
        if(root<0 .or. rank==root) then
          ! if root >= 0: only root gathers prefix sum
          blkweights_ps(0) = 0.0_r8k
        else
          ! if root >= 0: other procs do not receive blkweights_ps, set receive buffer to dummy
          blkweights_ps => dummy
        end if
        call fd4_part_sfc_exch_weights_r8k(domain, bal%sfcd_blkweights_ps(1), blkweights_ps(0), err, root)
        FD4_L1_ASSERT(err/=0,'fd4_part_sfc_exch_weights_r8k failed')

        if(domain%stat_enabled) call gettime_ofday(timing_us0)
    
      else
        !! allgather of weights + serial prefix sum of block weights

        if(domain%stat_enabled) call gettime_ofday(timing_us0)

        ! allocate array for sending the block weights
        if( size(mpi%sendreal,1) < domain%blocks%length ) then
          deallocate(mpi%sendreal)
          allocate(mpi%sendreal(domain%blocks%length))
        end if

        ! get the local block weights and write them to mpi%sendreal (ordered by SFC)
        iter%cur => rbtree_fd4_block_iter_first(iter%biter, domain%blocks)
        do while(associated(iter%cur))
          call fd4_part_sfc_c2i_qsz(domain, sfctype, iter%cur%pos, ib)
          mpi%sendreal(ib-bal%sfc_pstart+1) = iter%cur%weight
          iter%cur => rbtree_fd4_block_iter_next(iter%biter)
        end do

        if(domain%stat_enabled) then
          call gettime_ofday(timing_us1)
          domain%stat%values(bal%stat_pfxs_us) = domain%stat%values(bal%stat_pfxs_us) + (timing_us1-timing_us0)
        end if

        ! collect the block weights to mpi%recvreal (rwk array)
        call fd4_part_sfc_exch_weights(domain, err, mpi%sendreal, .false.)
        FD4_L1_ASSERT(err/=0,'fd4_part_sfc_exch_weights failed')

        if(domain%stat_enabled) call gettime_ofday(timing_us0)

        ! apply prefix sum to weight array
        maxweight = 0.0_r8k
        blkweights_ps(0) = 0.0_r8k
        do ib=1,nblocks
          blkweights_ps(ib) = REAL(mpi%recvreal(ib), r8k) ! convert to r8k
          maxweight = MAX(maxweight, blkweights_ps(ib))
          blkweights_ps(ib) = blkweights_ps(ib) + blkweights_ps(ib-1)
        end do

        if(domain%stat_enabled) then
          call gettime_ofday(timing_us1)
          domain%stat%values(bal%stat_pfxs_us) = domain%stat%values(bal%stat_pfxs_us) + (timing_us1-timing_us0)
          timing_us0 = timing_us1
        end if

      end if ! do_par_pfxs

      ! save old local partition
      oldp(1) = bal%sfc_pstart
      oldp(2) = oldp(1) + domain%blocks%length - 1
      
      ! save old partition vector
      if(.not. blkmap_update) then
        if(.not. associated(bal%partition_old)) then
          allocate(bal%partition_old(0:domain%mpi%nprocs), stat=err)
          FD4_L1_ASSERT(err/=0,'allocate bal%partition_old')
        end if
        bal%partition => bal%partition_old
        bal%partition_old => partition
        partition => bal%partition
      end if

    elseif(domain%blockcount == 0) then
      !! no blocks allocated yet
      
      maxweight = 0.0_r8k
      oldp(1:2) = -1

      if(root<0 .or. rank==root) then  ! if root >= 0: only root computes prefix sum

        if(domain%stat_enabled) call gettime_ofday(timing_us0)
      
        ! loop over all (non-allocated) blocks and fill blkweights_ps (prefix sum of weight array)
        blkweights_ps(0) = 0.0_r8k
        do ib=1,nblocks
          call fd4_part_sfc_i2c_qsz(domain, sfctype, ib, pos)
          blkweights_ps(ib) =   REAL( bal%defaultweight, r8k)  &
                              * REAL( PRODUCT( fd4_domain_blext(domain,pos) ), r8k )
          maxweight = MAX(maxweight, blkweights_ps(ib))
          blkweights_ps(ib) = blkweights_ps(ib) + blkweights_ps(ib-1)
        end do

        if(domain%stat_enabled) then
          call gettime_ofday(timing_us1)
          domain%stat%values(bal%stat_pfxs_us) = domain%stat%values(bal%stat_pfxs_us) + (timing_us1-timing_us0)
        end if

      end if
      
      if(domain%stat_enabled) call gettime_ofday(timing_us0)

    else
      FD4_L1_ERROR('not allowed in adaptive block mode')
    end if
    
    bal%last_sfctype = sfctype

    if(root<0 .or. rank==root) then  ! if root >= 0: only root computes
    
      if(get_maxweight/=0) then
        !accuracy=MAX(MIN(accuracy,1.0_r8k),0.1_r8k)
        !binsteps=MAX(MIN(binsteps,10000),1)
        ! compute the bottleneck error value which is necessary to get
        !  (accuracy * optimal_balance)
        ! as the resulting loadbalance of the 1d partitioning
        !  (eps is 0 if accuracy is 1.0: in this case the methods is exact if binsteps is large enough)
        eps = (1.0_r8k-accuracy) * blkweights_ps(nblocks) / (accuracy*nprocs)
        ! this would be the necessary number of binary search steps for the given accuracy:
        !steps = log2i( CEILING( maxweight / eps, i_k ) )
        if(get_maxweight==2) then
          ! maxweight was not determined in parallel with MPI reduce, so we need to do this now sequentially
          maxweight=blkweights_ps(1)-blkweights_ps(0)
          do ib=2,nblocks
            maxweight = MAX(blkweights_ps(ib)-blkweights_ps(ib-1), maxweight)
          end do
        end if
      end if

      !! call partitioning routine

      if(method==FD4_PART_SFC_BINSRCH) then
        ! good serial method
        call fd4_part_1d_binsrch(nblocks, nprocs, blkweights_ps(0:nblocks), maxweight, eps, binsteps, &
                                         partition, maxload, reqsteps)
        if(domain%stat_enabled) then
          domain%stat%values(bal%stat_binsteps) = MAX(domain%stat%values(bal%stat_binsteps), INT(reqsteps,i8k))
        end if
        !
      elseif(method==FD4_PART_SFC_PARALLEL) then    
        ! parallel method
        call fd4_part_1d_parallel(nblocks, nprocs, blkweights_ps(0:nblocks), maxweight, eps, binsteps, &
                                          mpi, partition, maxload, reqsteps, mpi_time, err)
        FD4_L1_ASSERT(err/=0,'fd4_part_1d_parallel failed')
        err = -1
        if(domain%stat_enabled) then
          domain%stat%values(bal%stat_binsteps) = MAX(domain%stat%values(bal%stat_binsteps), INT(reqsteps,i8k))
          domain%stat%values(bal%stat_mpi_us)   = domain%stat%values(bal%stat_mpi_us) + mpi_time
        end if
        !
      elseif(method==FD4_PART_SFC_RB) then
        ! recursive bisection
        call fd4_part_1d_simple_rb(nblocks, nprocs, blkweights_ps(0:nblocks), partition)
        maxload = blkweights_ps(nblocks) ! forces weird value for balance
        !
      elseif(method==FD4_PART_SFC_EBS) then
        ! exact bisection from pinar
        call fd4_part_1d_binsrch_ebs(nblocks, nprocs, blkweights_ps(0:nblocks), maxweight, eps, binsteps, .true., &
                                     partition, maxload, reqsteps)
        if(domain%stat_enabled) then
          domain%stat%values(bal%stat_binsteps) = MAX(domain%stat%values(bal%stat_binsteps), INT(reqsteps,i8k))
        end if
        !
      elseif(method==FD4_PART_SFC_SIMPLE) then
        ! simple method 1
        call fd4_part_1d_simple_h1(nblocks, nprocs, blkweights_ps(0:nblocks), partition, maxload)
        !
      else
        ! simple method 2
        call fd4_part_1d_simple_h2(nblocks, nprocs, blkweights_ps(0:nblocks), partition, maxload)
        !
      end if

      balance = REAL( blkweights_ps(nblocks) / (nprocs * maxload), r4k)
      
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_p1d_us) = domain%stat%values(bal%stat_p1d_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if

    end if
    

    if(root >= 0) then  ! if root >= 0: bcast partition vector to all
      partition(0) = 1
      partition(nprocs) = nblocks + 1
      call MPI_Bcast(partition(1), nprocs-1, MPI_I_K, root, mpi%comm, err)
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0)
        domain%stat%values(bal%stat_dspv_us) = domain%stat%values(bal%stat_dspv_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if
    end if

    !! finally apply partition to FD4

    err = -1

    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now

    ! allocate local list of blocks
    pstart = partition(rank)
    pend   = partition(rank+1)-1
    bal%sfc_pstart = pstart
    if(pstart<=pend) then
      if(root<0) then
        bal%localweight = blkweights_ps(pend) - blkweights_ps(pstart-1)
      else
        bal%localweight = 0.0_r8k
      end if
      call stack3_create(local_blocks, pend - pstart + 1)
      local_blocks%pos = pend - pstart + 1
    else
      bal%localweight = 0.0_r8k
      local_blocks%pos = 0
    end if

    call fd4_part_sfc_iter_init(sfciter, domain, sfctype)

    ! set local list of blocks
    if(pstart<=pend) then
      call fd4_part_sfc_iter_set(sfciter, pstart)
      do while(sfciter%idx>0 .and. sfciter%idx<=pend)
        local_blocks%a(1:3,sfciter%idx-pstart+1) = sfciter%p(1:3)
        call fd4_part_sfc_iter_next(sfciter)
      end do
    end if

    bal%blkmap_needs_update = .true.
    
    ! here is the actual blkmap update
    if(blkmap_update) then
      ! full update
      call fd4_part_sfc_update_blkmap(domain)
    else
      ! update only the necessary blocks
      call fd4_part_sfc_update_blkmap_loc(domain, sfciter, oldp, pstart, pend, local_blocks, err)
      FD4_L3_ASSERT(err/=0,'fd4_part_sfc_update_blkmap_loc failed')
      err = -1
    end if


    if(domain%stat_enabled) then
      call gettime_ofday(timing_us1)
      domain%stat%values(bal%stat_blkm_us) = domain%stat%values(bal%stat_blkm_us) + (timing_us1-timing_us0)
      ! fill statistics
      domain%stat%sfcbnd(1) = MIN(domain%stat%sfcbnd(1), pstart)
      domain%stat%sfcbnd(2) = MAX(domain%stat%sfcbnd(2), pend)
    end if

#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6,E13.4)') FD4_LINEHD,'new partition: ',local_blocks%pos, bal%localweight
#endif
  
    err = 0

  end subroutine fd4_part_sfc_weight2
  

  !! Create a new partition using block position. Considers block weights from domain%balance%blkweights.
  !!
  !! In adaptive block mode $all_blocks may be $NULL() and $all_blocks_len is ignored.
  !!
  !! If $blocks_changed is set to false: $fd4_part_sfc_weight reuses the hilbert indexes
  !! and the ranking of blocks by the hilbert index from the last call!
  !!
  !! THIS IS LEGACY CODE, use fd4_part_sfc_weight2 if not using adaptive block mode
  !!
  subroutine fd4_part_sfc_weight(domain, all_blocks, all_blocks_len, local_blocks, method, sfctype, &
                                 blocks_changed, binsteps, accuracy, balance, err)
    type(fd4_domain), intent(inout) :: domain            !! the domain
    integer(i_k)    , intent(in)    :: all_blocks_len    !! length of $all_blocks(3,:)
    integer(i_k)    , pointer       :: all_blocks(:,:)   !! array of positions of the blocks in the new partion
    type(stack3)    , intent(inout) :: local_blocks      !! stack3 of new local partition (differs from rank to rank)
    integer(i_k)    , intent(in)    :: method            !! 1D partitioning method to use
    integer(i_k)    , intent(in)    :: sfctype           !! space-filling curve to use
    logical         , intent(in)    :: blocks_changed    !! tell me if the blocks have changed (created/removed)
    integer(i_k)    , intent(in)    :: binsteps          !! max. number of steps for binary search methods
    !! target ratio from best possible load balance for binary steps estimation, *must be <= 1.0*
    real(r8k)       , intent(in)    :: accuracy
    real(r4k)       , intent(out)   :: balance           !! estimated balance of new partitioning
    integer(i_k)    , intent(out)   :: err               !! error status: 0...ok
    target :: domain
    
    integer :: rank, nprocs, nblocks
    integer(i_k) :: ib, i, j, k, crank, pos(3), pstart, pend, reqsteps
    real(r8k) :: maxweight, weight, maxload, eps
    real(r8k), pointer :: blkweights_ps(:)     ! prefix sum of sfc sorted block weights
    integer(i_k), pointer :: partition(:)      ! partition vector (start indices)
    integer(i8k) :: mpi_time, timing_us0, timing_us1
    type(fd4_domain_balance), pointer :: bal
    
    err = -1
    bal    => domain%balance
    rank   = domain%mpi%rank
    nprocs = domain%mpi%nprocs
    
    if(domain%varinfo%vthres_mode) then
      nblocks = all_blocks_len
    else
      ! ignore all_blocks and all_blocks_len in non-adaptive block mode
      nblocks = PRODUCT(domain%blnum(1:3))
    end if
    
    if(.not. associated(bal%blkweights)) then
      allocate(bal%blkweights(1:domain%blnum(1),1:domain%blnum(2),1:domain%blnum(3)), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate bal%blkweights')
    end if

    if(.not. associated(bal%blkweights_ps)) then
      allocate(bal%blkweights_ps(0:PRODUCT(domain%blnum(1:3))), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate bal%blkweights_ps')
    else
      if( ubound(bal%blkweights_ps,1) < PRODUCT(domain%blnum(1:3)) ) then
        deallocate(bal%blkweights_ps)
        allocate(bal%blkweights_ps(0:PRODUCT(domain%blnum(1:3))), stat=err)
        FD4_L1_ASSERT(err/=0,'allocate bal%blkweights_ps')
      end if
    end if
    blkweights_ps => bal%blkweights_ps

    if(.not. associated(bal%partition)) then
      allocate(bal%partition(0:domain%mpi%nprocs), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate bal%partition')
    end if
    partition => bal%partition

    ! collect the block weights to bal%blkweights
    call fd4_part_sfc_exch_weights(domain, err)
    FD4_L1_ASSERT(err/=0,'fd4_part_sfc_exch_weights failed')

    
    if(blocks_changed .or. sfctype /= bal%last_sfctype ) then
    
      ! check if bal%hidx and bal%rnk are large enough, if not, deallocate
      if(associated(bal%hidx)) then
        if(ubound(bal%hidx,1)<nblocks) then
          deallocate(bal%hidx)
          FD4_L3_ASSERT(.not.associated(bal%rnk),'rnk not associated')
          deallocate(bal%rnk)
        end if
      end if

      ! check if bal%hidx and bal%rnk are allocated
      if(.not. associated(bal%hidx)) then
        allocate(bal%hidx(nblocks))
        allocate(bal%rnk(nblocks))
      end if

      ! get index of each block on the space-filling curve
      if(domain%varinfo%vthres_mode) then
        ! adaptive block mode: loop over specified blocks
        do ib=1,all_blocks_len
          call fd4_part_sfc_c2i(domain, sfctype, all_blocks(1:3,ib), bal%hidx(ib))
        end do
      else
        ! non-adaptive block mode: loop over all blocks
        ib = 0
        do k=1,domain%blnum(3)
          pos(3) = k
          do j=1,domain%blnum(2)
            pos(2) = j
            do i=1,domain%blnum(1)
              pos(1) = i
              ib = ib + 1
              call fd4_part_sfc_c2i(domain, sfctype, pos, bal%hidx(ib))
            end do
          end do
        end do
      end if

      ! rank the indices on the Hilbert curve, so that hidx(rnk(:)) is sorted
      call mrgrnk(bal%hidx(1:nblocks), bal%rnk(1:nblocks))
      
      bal%last_sfctype = sfctype

    end if
    
    if(domain%stat_enabled) call gettime_ofday(timing_us0)

    ! apply prefix sum to weight array
    maxweight = 0.0
    blkweights_ps(0) = 0.0
    do ib=1,nblocks
      if(domain%varinfo%vthres_mode) then
        ! all_blocks present
        pos(1:3) = all_blocks(1:3,bal%rnk(ib))
      else
        ! compute block position from index in rnk
        pos(1:3) = fd4_domain_i2ccf(domain,bal%rnk(ib))
      end if
      weight = REAL( bal%blkweights( pos(1), pos(2), pos(3) ), r8k)
      blkweights_ps(ib) = weight + blkweights_ps(ib-1)
      maxweight = MAX(maxweight, weight)
    end do

    if(domain%stat_enabled) then
      call gettime_ofday(timing_us1)
      domain%stat%values(bal%stat_pfxs_us) = domain%stat%values(bal%stat_pfxs_us) + (timing_us1-timing_us0)
      timing_us0 = timing_us1
    end if

    if(method==FD4_PART_SFC_BINSRCH .or. method==FD4_PART_SFC_PARALLEL .or. method==FD4_PART_SFC_EBS) then
      !accuracy=MAX(MIN(accuracy,1.0_r8k),0.1_r8k)
      !binsteps=MAX(MIN(binsteps,10000),1)
      ! compute the bottleneck error value which is necessary to get
      !  (accuracy * optimal_balance)
      ! as the resulting loadbalance of the 1d partitioning
      !  (eps is 0 if accuracy is 1.0: in this case the methods is exact if binsteps is large enough)
      eps = (1.0_r8k-accuracy) * blkweights_ps(nblocks) / (accuracy*nprocs)
      ! this would be the necessary number of binary search steps for the given accuracy:
      !steps = log2i( CEILING( maxweight / eps, i_k ) )
    end if
    
    !! call partitioning routine

    if(method==FD4_PART_SFC_BINSRCH) then
      ! good serial method
      call fd4_part_1d_binsrch(nblocks, nprocs, blkweights_ps(0:nblocks), maxweight, eps, binsteps, &
                                       partition, maxload, reqsteps)
      if(domain%stat_enabled) then
        domain%stat%values(bal%stat_binsteps) = MAX(domain%stat%values(bal%stat_binsteps), INT(reqsteps,i8k))
      end if
      !
    elseif(method==FD4_PART_SFC_PARALLEL) then    
      ! parallel method
      call fd4_part_1d_parallel(nblocks, nprocs, blkweights_ps(0:nblocks), maxweight, eps, binsteps, &
                                        domain%mpi, partition, maxload, reqsteps, mpi_time, err)
      if(domain%stat_enabled) then
        domain%stat%values(bal%stat_binsteps) = MAX(domain%stat%values(bal%stat_binsteps), INT(reqsteps,i8k))
        domain%stat%values(bal%stat_mpi_us)   = domain%stat%values(bal%stat_mpi_us) + mpi_time
      end if
      !
    elseif(method==FD4_PART_SFC_RB) then
      ! recursive bisection
      call fd4_part_1d_simple_rb(nblocks, nprocs, blkweights_ps(0:nblocks), partition)
      maxload = blkweights_ps(nblocks) ! forces weird value for balance
      !
    elseif(method==FD4_PART_SFC_EBS) then
      ! exact bisection from pinar
      call fd4_part_1d_binsrch_ebs(nblocks, nprocs, blkweights_ps(0:nblocks), maxweight, eps, binsteps, .true., &
                                   partition, maxload, reqsteps)
      if(domain%stat_enabled) then
        domain%stat%values(bal%stat_binsteps) = MAX(domain%stat%values(bal%stat_binsteps), INT(reqsteps,i8k))
      end if
      !
    elseif(method==FD4_PART_SFC_SIMPLE) then
      ! simple method 1
      call fd4_part_1d_simple_h1(nblocks, nprocs, blkweights_ps(0:nblocks), partition, maxload)
      !
    else
      ! simple method 2
      call fd4_part_1d_simple_h2(nblocks, nprocs, blkweights_ps(0:nblocks), partition, maxload)
      !
    end if

    if(domain%stat_enabled) then
      call gettime_ofday(timing_us1)
      domain%stat%values(bal%stat_p1d_us) = domain%stat%values(bal%stat_p1d_us) + (timing_us1-timing_us0)
      timing_us0 = timing_us1
    end if

    balance = REAL( blkweights_ps(nblocks) / (nprocs * maxload), r4k)

    !! finally apply partition to FD4

    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now
    if(nblocks<PRODUCT(domain%blnum)) domain%blkmap(:,:,:,domain%blkmap_now) = -1
    ! set block map
    if(domain%varinfo%vthres_mode) then
      do crank=0,nprocs-1
        do ib=partition(crank),partition(crank+1)-1
          pos(1:3) = all_blocks(1:3,bal%rnk(ib))
          domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = crank
        end do
      end do
    else
      do crank=0,nprocs-1
        do ib=partition(crank),partition(crank+1)-1
          pos(1:3) = fd4_domain_i2ccf(domain,bal%rnk(ib))
          domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = crank
        end do
      end do
    end if
    ! set local list of blocks
    pstart = partition(rank)
    pend   = partition(rank+1)-1
    bal%sfc_pstart = pstart
    if(pstart<=pend) then
      bal%localweight = blkweights_ps(pend) - blkweights_ps(pstart-1)
      call stack3_create(local_blocks, pend - pstart + 1)
      local_blocks%pos = pend - pstart + 1
      if(domain%varinfo%vthres_mode) then
        do ib=1,local_blocks%pos
          local_blocks%a(1:3,ib) = all_blocks(1:3,bal%rnk(ib+pstart-1))
        end do
      else
        do ib=1,local_blocks%pos
          local_blocks%a(1:3,ib) = fd4_domain_i2ccf(domain,bal%rnk(ib+pstart-1))
        end do
      end if
    else
      bal%localweight = 0.0_r8k
      local_blocks%pos = 0
    end if
    
    ! apply periodic boundary on blkmap (which is allocated with a boundary to make life easier later on)
    call fd4_part_sfc_blkmap_boundary(domain)    
    
    if(domain%stat_enabled) then
      call gettime_ofday(timing_us1)
      domain%stat%values(bal%stat_blkm_us) = domain%stat%values(bal%stat_blkm_us) + (timing_us1-timing_us0)
      ! fill statistics
      domain%stat%sfcbnd(1) = MIN(domain%stat%sfcbnd(1), pstart)
      domain%stat%sfcbnd(2) = MAX(domain%stat%sfcbnd(2), pend)
    end if
    
#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6,E13.4)') FD4_LINEHD,'new partition: ',local_blocks%pos, bal%localweight
#endif
   
  
    err = 0

  end subroutine fd4_part_sfc_weight



  !! Exchange the weights of my blocks with all other processes.
  !!
  !! difference to fd4_part_sfc_exch_weights:
  !! - no legacy stuff
  !! - r8k weights only
  !!
  !! Internal routine.
  subroutine fd4_part_sfc_exch_weights_r8k(domain, localweights, globalweights, err, opt_root)
    type(fd4_domain), intent(inout), target :: domain         !! the domain
    real(r8k),        intent(in)            :: localweights   !! start of array of local block weights
    real(r8k),        intent(inout)         :: globalweights  !! (start-1) of output array of global block weights
    integer(i_k),     intent(out)           :: err            !! error status: 0...ok
    integer(i_k),     intent(in)            :: opt_root       !! only gather weights in rank $opt_root, no allgather
    optional :: opt_root
    
    type(fd4_mpi), pointer :: mpi
    integer(i_k) :: crank, root
    integer(i8k) :: timing_us0, timing_us1, timing_us_entry
    
    root = -1
    if(present(opt_root)) root = opt_root

    mpi => domain%mpi
    
    call fd4_mpi_realloc_array(mpi%recvcnt, mpi%nprocs, 0, err)
    FD4_L1_ASSERT(err/=0,'fd4_mpi_realloc_array failed')
    
    err = -1
    
    if(domain%stat_enabled) call gettime_ofday(timing_us_entry)

    ! loop over partition vector (and perform inverse prefix sum)
    do crank=0,mpi%nprocs-1
      mpi%recvcnt(crank) = domain%balance%partition(crank+1) - domain%balance%partition(crank)
    end do

    FD4_L3_ASSERT(mpi%recvcnt(mpi%rank)/=domain%blocks%length,'nblocks not as expected')
    
    if(domain%stat_enabled) call gettime_ofday(timing_us0)

    ! exchange the block weights
    if(root>=0) then
      call MPI_Gatherv (localweights,  mpi%recvcnt(mpi%rank),    MPI_R8K, &
                        globalweights, mpi%recvcnt(0),  domain%balance%partition(0), MPI_R8K, root, mpi%comm, err )
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Gatherv failed')
    else
      call MPI_Allgatherv (localweights,  mpi%recvcnt(mpi%rank),    MPI_R8K, &
                           globalweights, mpi%recvcnt(0),  domain%balance%partition(0), MPI_R8K, mpi%comm, err )
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Allgatherv failed')
    end if

    if(domain%stat_enabled) then
      call gettime_ofday(timing_us1)
      domain%stat%values(domain%balance%stat_mpi_us) = domain%stat%values(domain%balance%stat_mpi_us) + (timing_us1-timing_us0)
      domain%stat%values(domain%balance%stat_exwgt_us) = domain%stat%values(domain%balance%stat_exwgt_us) &
                                                       + (timing_us1-timing_us_entry)
    end if

    err = 0
    
  end subroutine fd4_part_sfc_exch_weights_r8k



  !! Exchange the weights of my blocks with all other processes.
  !!
  !! Updates domain%balance%blkweights.
  !!
  !! Internal routine.
  subroutine fd4_part_sfc_exch_weights(domain, err, opt_localweights, opt_set_blkweights)
    type(fd4_domain), intent(inout), target :: domain       !! the domain
    integer(i_k),     intent(out)           :: err          !! error status: 0...ok
    real(rwk),        intent(in),optional   :: opt_localweights(:) !! optional array of local block weights
    logical,          intent(in),optional   :: opt_set_blkweights  !! set $domain%$blkweights? default: yes, if allocated
    
    type(fd4_mpi), pointer :: mpi
    integer(i_k) :: crank, nblocks, k, j, i
    type(fd4_iter) :: iter
    logical :: set_blkweights
    integer(i8k) :: timing_us0, timing_us1, timing_us_entry

    mpi => domain%mpi
    
    call fd4_mpi_realloc_array(mpi%recvcnt, mpi%nprocs, 0, err)
    FD4_L1_ASSERT(err/=0,'fd4_mpi_realloc_array failed')
    call fd4_mpi_realloc_array(mpi%recvdis, mpi%nprocs, 0, err)
    FD4_L1_ASSERT(err/=0,'fd4_mpi_realloc_array failed')
    
    nblocks = 0
    err = -1
    
    if(domain%stat_enabled) call gettime_ofday(timing_us_entry)

    set_blkweights = .true.
    if(present(opt_set_blkweights)) set_blkweights = opt_set_blkweights
    
    ! create recv counts for MPI_Allgatherv (number of blocks of each process)
    if(domain%balance%last_sfctype>0 .and. associated(domain%balance%partition)) then
      ! loop over partition vector (and perform inverse prefix sum)
      do crank=0,mpi%nprocs-1
        mpi%recvcnt(crank) = domain%balance%partition(crank+1) - domain%balance%partition(crank)
      end do
    else
      FD4_L3_ASSERT(domain%balance%blkmap_needs_update,'blkmap needs update')
      mpi%recvcnt(:) = 0
      ! loop over blkmap
      do k=1,domain%blnum(3)
        do j=1,domain%blnum(2)
          do i=1,domain%blnum(1)
            crank = domain%blkmap(i,j,k,domain%blkmap_now)
            if( crank >= 0 ) then
              mpi%recvcnt(crank) = mpi%recvcnt(crank) + 1
            end if
          end do
        end do
      end do
    end if

    FD4_L3_ASSERT(mpi%recvcnt(mpi%rank)/=domain%blocks%length,'nblocks not as expected')
    
    ! build the recv displacements
    mpi%recvdis(0) = 0
    do crank=1,mpi%nprocs-1
      mpi%recvdis(crank) = mpi%recvcnt(crank-1) + mpi%recvdis(crank-1)
    end do

    nblocks = mpi%recvdis(mpi%nprocs-1) + mpi%recvcnt(mpi%nprocs-1)

    ! this may fail, because domain%blockcount may be inconsistent
    ! after calling fd4_util_put_array
    !FD4_L3_ASSERT(nblocks/=domain%blockcount,'nblocks not as expected')
    
    ! allocate array for receiving the block weights
    if( size(mpi%recvreal,1) < nblocks ) then
      deallocate(mpi%recvreal)
      allocate(mpi%recvreal(nblocks))
    end if

    if(present(opt_localweights)) then
      !! use subroutine argument as local weights
      
      FD4_L3_ASSERT(size(opt_localweights)<mpi%recvcnt(mpi%rank),'too few weights')

      ! exchange the block weights
      if(domain%stat_enabled) call gettime_ofday(timing_us0)
      call MPI_Allgatherv (opt_localweights(1), mpi%recvcnt(mpi%rank),                 MPI_RWK, &
                           mpi%recvreal(1),     mpi%recvcnt(0),        mpi%recvdis(0), MPI_RWK, mpi%comm, err )
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Allgatherv failed')
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(domain%balance%stat_mpi_us) = domain%stat%values(domain%balance%stat_mpi_us) + (timing_us1-timing_us0)
      end if
      
    else
      !! fill mpi%sendreal with the local weights

      ! allocate array for sending the block weights
      if( size(mpi%sendreal,1) < mpi%recvcnt(mpi%rank) ) then
        deallocate(mpi%sendreal)
        allocate(mpi%sendreal(mpi%recvcnt(mpi%rank)))
      end if

      ! get the local block weights and write them to mpi%sendreal
      i = 0
      iter%cur => rbtree_fd4_block_iter_first(iter%biter, domain%blocks)
      do while(associated(iter%cur))
        i = i + 1
        mpi%sendreal(i) = iter%cur%weight
        iter%cur => rbtree_fd4_block_iter_next(iter%biter)
      end do
    
      FD4_L3_ASSERT(i/=domain%blocks%length,'nblocks not as expected')
      
      ! exchange the block weights
      if(domain%stat_enabled) call gettime_ofday(timing_us0)
      call MPI_Allgatherv (mpi%sendreal(1), mpi%recvcnt(mpi%rank),                 MPI_RWK, &
                           mpi%recvreal(1), mpi%recvcnt(0),        mpi%recvdis(0), MPI_RWK, mpi%comm, err )
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Allgatherv failed')
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(domain%balance%stat_mpi_us) = domain%stat%values(domain%balance%stat_mpi_us) + (timing_us1-timing_us0)
      end if
    
    end if
   
    
    if(set_blkweights .and. associated(domain%balance%blkweights)) then
      ! for legacy sfc method only
      ! fd4_part_sfc_weight2 does not allocate domain%balance%blkweights
      
      err = -1
      FD4_L3_ASSERT(domain%balance%blkmap_needs_update,'blkmap needs update')

      ! loop over blkmap and fill domain%balance%blkweights
      ! this method assumes the loop order in this loop has the same ordering
      ! as the index used to sort the blocks in domain%blocks
      do k=1,domain%blnum(3)
        do j=1,domain%blnum(2)
          do i=1,domain%blnum(1)
            crank = domain%blkmap(i,j,k,domain%blkmap_now)
            if( crank >= 0 ) then
              domain%balance%blkweights(i,j,k) = mpi%recvreal( mpi%recvdis(crank)+1 )
              mpi%recvdis(crank) = mpi%recvdis(crank) + 1
            else
              ! set default weight
              !TODO: do this only if the block has been removed last load balancing
              !      (or block will be added this time)
              domain%balance%blkweights(i,j,k) = domain%balance%defaultweight  &
               * REAL( PRODUCT( fd4_domain_blext_ijk(domain,i,j,k) ), rwk )
            end if
          end do
        end do
      end do

#if ( FD4_VERBOSE_LEVEL >= 3 )
      do crank=1,mpi%nprocs-1
        if(mpi%recvdis(crank) - mpi%recvdis(crank-1) /= mpi%recvcnt(crank)) then
          FD4_L1_ERROR('inconsistency after loop over blkmap')
        end if
      end do
#endif

    end if ! associated(domain%balance%blkweights)
    
    if(domain%stat_enabled) then
      call gettime_ofday(timing_us1)
      domain%stat%values(domain%balance%stat_exwgt_us) = domain%stat%values(domain%balance%stat_exwgt_us) &
                                                       + (timing_us1-timing_us_entry)
    end if

    err = 0
    
  end subroutine fd4_part_sfc_exch_weights
  



  !! Create a new partition using block position. Weights are not considered.
  !!
  !! The special option $add_only effects:
  !! - all existing blocks (in blkmap) are left where they are
  !! - new blocks are added as if the option was not set
  !! - this (intentionally) results in a non balanced partitioning, but minimizes transfer costs
  !!   when - in a second partitioning step - blocks are migrated
  !!
  subroutine fd4_part_sfc_noweight(domain, all_blocks, all_blocks_len, local_blocks, sfctype, add_only, opt_blocks_changed)
    type(fd4_domain), intent(inout) :: domain                         !! the domain
    integer(i_k)    , intent(in)    :: all_blocks_len                 !! length of $all_blocks(3,:)
    integer(i_k)    , intent(in)    :: all_blocks(:,:)                !! array of positions of the blocks in the new partion
    type(stack3)    , intent(inout) :: local_blocks                   !! stack3 of new local partition (differs from rank to rank)
    integer(i_k)    , intent(in)    :: sfctype                        !! space-filling curve to use
    logical,optional, intent(in)    :: add_only                       !! only add *new* blocks, but don't touch existing blocks
    logical,optional, intent(in)    :: opt_blocks_changed             !! tell me if the blocks have changed (created/removed)
    
    integer :: rank, nprocs
    integer(i_k) :: i, crank, psum, pos(3)
    logical :: ao
    logical :: blocks_changed
    
    FD4_L1_ASSERT(all_blocks_len==0,'need allblocks')
    
    if(present(add_only)) then
      ao = add_only
    else
      ao = .false.
    end if
    
    rank   = domain%mpi%rank
    nprocs = domain%mpi%nprocs
    blocks_changed = .true.
    
    if(present(opt_blocks_changed)) blocks_changed = opt_blocks_changed
    
    ! check if domain%balance%hidx and domain%balance%rnk are large enough, if not, deallocate
    if(associated(domain%balance%hidx)) then
      if(ubound(domain%balance%hidx,1)<all_blocks_len) then
        deallocate(domain%balance%hidx)
        FD4_L3_ASSERT(.not.associated(domain%balance%rnk),'rnk not associated')
        deallocate(domain%balance%rnk)
      end if
    end if

    ! check if domain%balance%hidx and domain%balance%rnk are allocated
    if(.not. associated(domain%balance%hidx)) then
      allocate(domain%balance%hidx(all_blocks_len))
      allocate(domain%balance%rnk(all_blocks_len))
      blocks_changed = .true.
    end if

    if(blocks_changed .or. sfctype /= domain%balance%last_sfctype ) then

      ! get index of each block on the space-filling curve
      do i=1,all_blocks_len
        call fd4_part_sfc_c2i(domain, sfctype, all_blocks(1:3,i), domain%balance%hidx(i))
      end do

      ! rank the indices on the hilbert curve, so that hidx(rnk(:)) is sorted
      call mrgrnk(domain%balance%hidx(1:all_blocks_len), domain%balance%rnk(1:all_blocks_len))

      domain%balance%last_sfctype = sfctype

    end if
    
    if(ao) domain%balance%last_sfctype = -1
        
    
    call stack3_create(local_blocks, all_blocks_len/nprocs)
    crank = 0
    psum = 0
    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now
    if(all_blocks_len<PRODUCT(domain%blnum)) domain%blkmap(:,:,:,domain%blkmap_now) = -1
    domain%balance%sfc_pstart = -1
    do i=1,all_blocks_len
      psum = psum + nprocs
      if(all_blocks_len*(crank+1)<psum) crank=crank+1
      pos(1:3) = all_blocks(1:3,domain%balance%rnk(i))
      if(ao .and. domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_last ) /= -1 ) then
        if(domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_last ) == domain%mpi%rank ) then
          call stack3_push(local_blocks, pos(1:3))
        end if
        domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_last )
      else
        domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = crank
        if(rank == crank) then
          call stack3_push(local_blocks, pos(1:3))
          if(domain%balance%sfc_pstart<0) domain%balance%sfc_pstart = i
        end if
      end if
    end do
    
    ! apply periodic boundary on blkmap (which is allocated with a boundary to make life easier later on)
    call fd4_part_sfc_blkmap_boundary(domain)    
    
#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6)') FD4_LINEHD,'new partition: ',local_blocks%pos
#endif
  
  end subroutine fd4_part_sfc_noweight


  !! Update domain%blkmap from domain%balance%partition using the last used SFC
  !! for given subset of blocks.
  !!
  subroutine fd4_part_sfc_update_blkmap_bnd(domain, bbnd, err)
    type(fd4_domain), intent(inout) :: domain     !! the domain
    integer(i_k),     intent(in)    :: bbnd(3,2)  !! bounds in block map
    integer(i_k),     intent(out)   :: err        !! error status: 0...ok (only L3 checking)

    integer(i_k) :: crank, i, j, k, pos(3), idx, nbnd(3,2), dim
    integer(i_k), pointer :: partition(:)      ! partition vector (start indices)
    
    err = 0
    if(ANY( bbnd(:,1) > bbnd(:,2) )) return
    
    err = -1
    FD4_L3_ASSERT(domain%varinfo%vthres_mode,'not allowed in vthres_mode')
    FD4_L3_ASSERT(ANY(bbnd(:,1)<1),'bbnd(:,1)<1')
    FD4_L3_ASSERT(ANY(bbnd(:,2)>domain%blnum),'bbnd(:,2)>blnum')

    partition => domain%balance%partition
    FD4_L3_ASSERT(.not.associated(partition),'partition not allocated')

    FD4_L4_INFO('blkmap update '//p2s(bbnd(:,1))//' - '//p2s(bbnd(:,2)))

    do k=bbnd(3,1),bbnd(3,2)
      pos(3) = k
      do j=bbnd(2,1),bbnd(2,2)
        pos(2) = j
        do i=bbnd(1,1),bbnd(1,2)
          pos(1) = i
          call fd4_part_sfc_c2i_qsz(domain, domain%balance%last_sfctype, pos, idx)
          FD4_L3_ASSERT(idx<1,'idx < 1')
          FD4_L3_ASSERT(idx>=partition(domain%mpi%nprocs),'idx > nblocks')
          call fd4_part_sfc_find_partition(partition, idx, crank)
          FD4_L3_ASSERT(crank<0.or.crank>domain%mpi%nprocs,'crank out of bounds')
          domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = crank
        end do
      end do
    end do
    
    ! also update neighbor blocks
    !  consider the situation at the domain boundary:
    !  - peridodic boundaries: get block owner from opposite domain boundary
    !  - no periodic boundaries: set to -1
    do dim=1,3
      ! dir -1
      nbnd = bbnd
      nbnd(dim,1:2) = bbnd(dim,1) - 1
      if(bbnd(dim,1)>1 .or. domain%periodic(dim)) then
        call fd4_part_sfc_update_blkmap_loop(nbnd, dim, MOD( nbnd(dim,1) + domain%blnum(dim) - 1, domain%blnum(dim) ) + 1, err )
        FD4_L3_ASSERT(err/=0,'fd4_part_sfc_update_blkmap_loop failed')
      else
        !TODO: do we need this? maybe this is never touched and thus always -1
        !domain%blkmap( nbnd(1,1):nbnd(1,2), nbnd(2,1):nbnd(2,2), nbnd(3,1):nbnd(3,2), domain%blkmap_now ) = -1
#if ( FD4_VERBOSE_LEVEL >= 3 )
        if( ANY( domain%blkmap( nbnd(1,1):nbnd(1,2), nbnd(2,1):nbnd(2,2), nbnd(3,1):nbnd(3,2), domain%blkmap_now ) /= -1 ) ) then
          FD4_L1_ERROR('blkmap not -1!')
        end if
#endif
      end if
      ! dir +1
      nbnd = bbnd
      nbnd(dim,1:2) = bbnd(dim,1) + 1
      if(bbnd(dim,1)<domain%blnum(dim) .or. domain%periodic(dim)) then
        call fd4_part_sfc_update_blkmap_loop(nbnd, dim, MOD( nbnd(dim,1) + domain%blnum(dim) - 1, domain%blnum(dim) ) + 1, err )
        FD4_L3_ASSERT(err/=0,'fd4_part_sfc_update_blkmap_loop failed')
      else
        !TODO: do we need this? maybe this is never touched and thus always -1
        !domain%blkmap( nbnd(1,1):nbnd(1,2), nbnd(2,1):nbnd(2,2), nbnd(3,1):nbnd(3,2), domain%blkmap_now ) = -1
#if ( FD4_VERBOSE_LEVEL >= 3 )
        if( ANY( domain%blkmap( nbnd(1,1):nbnd(1,2), nbnd(2,1):nbnd(2,2), nbnd(3,1):nbnd(3,2), domain%blkmap_now ) /= -1 ) ) then
          FD4_L1_ERROR('blkmap not -1!')
        end if
#endif
      end if
    end do
    
    err = 0

  contains
  
    !! update blkmap for blocks specified by bbnd
    !! get the block owner from a different position at ipos(dim)=innerpos
    !! (in case of periodic boundaries)
    !!
    subroutine fd4_part_sfc_update_blkmap_loop(bbnd, dim, innerpos, err)
      integer(i_k),     intent(in)    :: bbnd(3,2)  !! bounds in block map
      integer(i_k),     intent(in)    :: dim, innerpos  !! dim and value of changed position
      integer(i_k),     intent(out)   :: err        !! error status: 0...ok (only L3 checking)
      integer(i_k) :: ipos(3)
      err = -1
      do k=bbnd(3,1),bbnd(3,2)
        pos(3) = k
        do j=bbnd(2,1),bbnd(2,2)
          pos(2) = j
          do i=bbnd(1,1),bbnd(1,2)
            pos(1) = i
            ipos = pos
            ipos(dim) = innerpos
            call fd4_part_sfc_c2i_qsz(domain, domain%balance%last_sfctype, ipos, idx)
            FD4_L3_ASSERT(idx<1,'idx < 1')
            FD4_L3_ASSERT(idx>=partition(domain%mpi%nprocs),'idx > nblocks')
            call fd4_part_sfc_find_partition(partition, idx, crank)
            FD4_L3_ASSERT(crank<0.or.crank>domain%mpi%nprocs,'crank out of bounds')
            domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = crank
          end do
        end do
      end do
      err = 0
    end subroutine fd4_part_sfc_update_blkmap_loop

  end subroutine fd4_part_sfc_update_blkmap_bnd


  !! Update only necessary parts of domain%blkmap from domain%balance%partition using the last used SFC
  !!
  !! Does not make a full update of the block map, only update:
  !!  - local blocks
  !!  - old local blocks
  !!  - neighbor blocks of local blocks
  !!  - blocks which intersect with local couple arrays
  !!  - local blocks in old block map using partition_old
  !!
  !! Some routines need the full blockmap (I/O stuff), they must call
  !! $fd4_part_sfc_update_blkmap before accessing the blkmap!
  !!
  subroutine fd4_part_sfc_update_blkmap_loc(domain, sfciter, oldp, pstart, pend, local_blocks, err)
    type(fd4_domain), intent(inout) :: domain        !! the domain
    type(part_sfc_iter), intent(inout) :: sfciter    !! sfc iterator correctly initialized for the domain
    integer(i_k)    , intent(inout) :: oldp(2)       !! start and end index of old own partition
    integer(i_k)    , intent(in)    :: pstart        !! start index of new own partition
    integer(i_k)    , intent(in)    :: pend          !! end index of new own partition
    type(stack3)    , intent(in)    :: local_blocks  !! stack3 of new local partition (differs from rank to rank)
    integer(i_k)    , intent(out)   :: err           !! error status: 0...ok
    target :: domain

    integer :: rank, nprocs, nblocks
    integer(i_k) :: ib, crank, pos(3)
    integer(i_k), pointer :: partition(:)      ! partition vector (start indices)
    integer(i_k), pointer :: partition_old(:)  ! old partition vector (start indices)
    !integer(i8k) :: timing_us0, timing_us1
    integer(i_k) :: dir, dim, idx, ppos(3)
    type(fd4_domain_balance), pointer :: bal
    
    err = -1

    bal => domain%balance
    partition => bal%partition
    rank = domain%mpi%rank
    nprocs = domain%mpi%nprocs
    nblocks = PRODUCT(domain%blnum(1:3))

    ! set blkmap to -1 only at places, where we will check for blkmap < 0 in the next lines
    !domain%blkmap(:,:,:,domain%blkmap_now) = -1
    ! position of neighbor blocks of local blocks
    do ib=1,local_blocks%pos
      do dim=1,3
        do dir=1,2
          pos(1:3) = local_blocks%a(1:3,ib)
          pos(dim) = pos(dim) + fd4_incdec(dir)
          domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = -1
        end do
      end do
    end do
    ! position of local couple arrays
    !domain%blkmap( domain%cplbnd(1,1):domain%cplbnd(1,2), &
    !               domain%cplbnd(2,1):domain%cplbnd(2,2), &
    !               domain%cplbnd(3,1):domain%cplbnd(3,2),   domain%blkmap_now) = -1

    !! now update selected sections of the blkmap

    ! local blocks
    do ib=1,local_blocks%pos
      domain%blkmap( local_blocks%a(1,ib), local_blocks%a(2,ib), local_blocks%a(3,ib), domain%blkmap_now ) = rank
    end do

    ! old local blocks
    if(oldp(1)>0 .and. oldp(1)<=oldp(2)) then
      crank = -1
      if(pstart<=pend) then
        if(pstart <= oldp(1) .and. oldp(1) <= pend) then
          oldp(1) = pend + 1
          crank = rank + 1
        end if
        if(pstart <= oldp(2) .and. oldp(2) <= pend) then
          oldp(2) = pstart - 1
        end if
      end if
      if(oldp(1)<=nblocks .and. oldp(1)<=oldp(2) .and. oldp(2)>=1) then
        FD4_L3_ASSERT(oldp(1)<1,'oldp(1) < 1 '//i2s(oldp(1)))
        FD4_L3_ASSERT(oldp(2)>nblocks,'oldp(2) > nblocks '//i2s(oldp(2)))
        if(crank < 0) then
          call fd4_part_sfc_find_partition(partition, oldp(1), crank)
        end if
        FD4_L3_ASSERT(crank<0.or.crank>nprocs,'crank out of bounds')
        call fd4_part_sfc_iter_set(sfciter, oldp(1))
        do while(sfciter%idx>0 .and. sfciter%idx<=oldp(2))
          ! advance rank
          do while( crank < nprocs-1 .and. sfciter%idx >= partition(crank+1) )
            crank = crank + 1
          end do
          domain%blkmap( sfciter%p(1), sfciter%p(2), sfciter%p(3), domain%blkmap_now ) = crank
          call fd4_part_sfc_iter_next(sfciter)
        end do
      end if
    end if

    ! neighbor blocks of local blocks
    do ib=1,local_blocks%pos
      do dim=1,3
        do dir=1,2
          pos(1:3) = local_blocks%a(1:3,ib)
          pos(dim) = pos(dim) + fd4_incdec(dir)
          if( domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) < 0 ) then
            ! need to update this block
            ! get pos within domain (periodic boundaries)
            ppos = pos
            where(domain%periodic) ppos = MOD( ppos + domain%blnum - 1, domain%blnum ) + 1
            ! skip this neigbor block if out of domain boundary and no periodic boundaries
            if( ANY( ppos<1 .or. ppos>domain%blnum ) ) cycle
            ! get sfc index
            call fd4_part_sfc_c2i_qsz(domain, sfciter%sfctype, ppos, idx)
            FD4_L3_ASSERT(idx<1,'idx < 1')
            FD4_L3_ASSERT(idx>nblocks,'idx > nblocks')
            call fd4_part_sfc_find_partition(partition, idx, crank)
            FD4_L3_ASSERT(crank<0.or.crank>nprocs,'crank out of bounds')
            domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = crank
          end if
        end do
      end do
    end do

    ! blocks which intersect with local couple arrays
    ! (do not check for blkmap < 0)
    call fd4_part_sfc_update_blkmap_bnd(domain, domain%cplbnd, err)
    FD4_L3_ASSERT(err/=0,'fd4_part_sfc_update_blkmap_bnd failed')
    err = -1

    ! local blocks in old block map
    if(associated(bal%partition_old) .and. pstart<=pend) then
      partition_old => bal%partition_old
      FD4_L3_ASSERT(pstart<1,'pstart < 1')
      FD4_L3_ASSERT(pstart>nblocks,'pstart > nblocks')
      call fd4_part_sfc_find_partition(partition_old, pstart, crank)
      FD4_L3_ASSERT(crank<0.or.crank>nprocs,'crank out of bounds')
      call fd4_part_sfc_iter_set(sfciter, pstart)
      do while(sfciter%idx>0 .and. sfciter%idx<=pend)
        ! advance rank
        do while( crank < nprocs-1 .and. sfciter%idx >= partition_old(crank+1) )
          crank = crank + 1
        end do
        domain%blkmap( sfciter%p(1), sfciter%p(2), sfciter%p(3), domain%blkmap_last ) = crank
        call fd4_part_sfc_iter_next(sfciter)
      end do
    end if
    
    err = 0
  
  end subroutine fd4_part_sfc_update_blkmap_loc


  !! Update complete domain%blkmap from domain%balance%partition using the last used SFC
  !!
  subroutine fd4_part_sfc_update_blkmap(domain)
    type(fd4_domain), intent(inout) :: domain     !! the domain
    target :: domain

    integer(i_k) :: crank
    integer(i_k), pointer :: partition(:)      ! partition vector (start indices)
    type(part_sfc_iter) :: sfciter
    
    if(.not. domain%balance%blkmap_needs_update) return
    
    if(domain%mpi%rank==0) FD4_L4_INFO('fd4_part_sfc_update_blkmap')

    partition => domain%balance%partition

    ! iterate over SFC to set block map and local partition
    crank = 0
    call fd4_part_sfc_iter_init(sfciter, domain, domain%balance%last_sfctype)
    do while(sfciter%idx>0)
      ! advance rank
      do while( crank < domain%mpi%nprocs-1 .and. sfciter%idx >= partition(crank+1) )
        crank = crank + 1
      end do
      ! set block map
      domain%blkmap( sfciter%p(1), sfciter%p(2), sfciter%p(3), domain%blkmap_now ) = crank
      call fd4_part_sfc_iter_next(sfciter)
    end do

    ! apply periodic boundary on blkmap (which is allocated with a boundary to make life easier later on)
    call fd4_part_sfc_blkmap_boundary(domain)
    
    domain%balance%blkmap_needs_update = .false.

    if(domain%stat_enabled) then
      domain%stat%values(domain%balance%stat_blkmap) = domain%stat%values(domain%balance%stat_blkmap) + 1
    end if

  end subroutine fd4_part_sfc_update_blkmap

end module fd4_part_sfc_mod
