! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2013 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                Statistics and Profiling of FD4               !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Internal statistics and Profiling of FD4
!!
!! User interface: Use $fd4_domain_dump_stats to dump the current values to stdout.
!!
!! Use $fd4_stat_add to create a new statistics.
!! All ranks *must* call this function in the same order, so that the
!! index returned by $fd4_stat_add is equal in all ranks.
!!
!! To set a value, just access it directly:
!!> type(fd4_stat) :: stat
!!  call fd4_stat_add(stat, 'Bytes transferred', bytes_idx)
!!  stat%values(bytes_idx) = size_bytes
!!<
!!
!! @see fd4_domain_mod
!! @author Matthias Lieber
module fd4_stat_mod

  use fd4_globaldef_mod
  use fd4_mpi_mod

  USE_MPI
  
  implicit none
  private

  INCLUDE_MPIFH
  
  !! max. length of a statistics name
  integer, parameter :: FD4_STATNAME_LEN = 20
  
  !! initial capacity of the statistics
  integer, parameter :: FD4_STAT_INITIAL_SIZE = 29
  
  !! logical unit for stdout
  integer, parameter :: FD4_STAT_STDOUT = 6
  
  public :: fd4_stat, FD4_STAT_STDOUT
  public :: fd4_stat_add, fd4_stat_dump, fd4_stat_reset, fd4_stat_free, fd4_stat_countup
  
  !! type for a set of statistics
  type fd4_stat
    ! run-time defined statistics
    integer(i8k), pointer :: values(:) => null()                !! values of the statistics
    character(FD4_STATNAME_LEN), pointer :: names(:) => null()  !! names of the statistics
    logical, pointer :: dumpifset(:) => null()                  !! dump the counter only when sum>0?
    integer(i_k) :: nstat = 0
    ! other statistics
    integer(i_k) :: sfcbnd(2) = (/HUGE(0_i_k),-1/)              !! min and max SFC index of all partitions
  end type
  
  
contains

  !! Add a new statistic to the $fd4_stat.
  !!
  !! The index $idx in the arrays $values and $names of the $fd4_stat is returned.
  !! If the name already exists, the corresponding index is returned.
  !!
  !! Returns negative $idx on error.
  !!
  subroutine fd4_stat_add(stat, name, idx, opt_dump_only_if_set)
    type(fd4_stat), intent(inout) :: stat       !! the statistics object
    character(*), intent(in) :: name            !! name of the new statistic
    integer(i_k), intent(out) :: idx            !! added index
    logical, intent(in), optional :: opt_dump_only_if_set !! dump the counter only when sum>0
    
    type(fd4_stat) :: tmp
    integer :: i
    
    idx = -1

    if(.not. associated(stat%values)) then
      ! initialize stat
      allocate(stat%values   (FD4_STAT_INITIAL_SIZE))
      allocate(stat%names    (FD4_STAT_INITIAL_SIZE))
      allocate(stat%dumpifset(FD4_STAT_INITIAL_SIZE))
      stat%values = 0_i8k
      stat%names = ''
      stat%dumpifset = .false.
      stat%nstat = 0
    end if
    
    FD4_L3_ASSERT(stat%nstat>size(stat%values),'nstat too large')

    ! check if the name already exists
    do i=1,stat%nstat
      if(STRIM(stat%names(i))==STRIM(name)) then
        idx = i
        FD4_L5_INFO('found statistics '//STRIM(name))
        exit
      end if
    end do
    
    if(idx<0) then
      ! name was not found
    
      if(stat%nstat==size(stat%values)) then
        ! enlarge stat
        allocate(tmp%values   (stat%nstat+FD4_STAT_INITIAL_SIZE))
        allocate(tmp%names    (stat%nstat+FD4_STAT_INITIAL_SIZE))
        allocate(tmp%dumpifset(stat%nstat+FD4_STAT_INITIAL_SIZE))
        tmp%values = 0_i8k
        tmp%names = ''
        tmp%dumpifset = .false.
        tmp%values   (1:stat%nstat) = stat%values   (1:stat%nstat)
        tmp%names    (1:stat%nstat) = stat%names    (1:stat%nstat)
        tmp%dumpifset(1:stat%nstat) = stat%dumpifset(1:stat%nstat)
        deallocate(stat%values)
        deallocate(stat%names)
        deallocate(stat%dumpifset)
        stat%values     => tmp%values
        stat%names      => tmp%names
        stat%dumpifset  => tmp%dumpifset
      end if

      ! add new statistic
      stat%nstat = stat%nstat + 1
      idx = stat%nstat
      i = MIN(len_trim(name),FD4_STATNAME_LEN)
      stat%names(stat%nstat)(1:i) = name(1:i)
      if(present(opt_dump_only_if_set)) stat%dumpifset(stat%nstat) = opt_dump_only_if_set
      FD4_L5_INFO('added statistics '//STRIM(name))

    end if
    
  end subroutine fd4_stat_add
  

  !! Reset all values to zero.
  !!
  subroutine fd4_stat_reset(stat)
    type(fd4_stat), intent(inout) :: stat

    if(associated(stat%values)) stat%values = 0_i8k
    
  end subroutine fd4_stat_reset


  !! Free memory of an $fd4_stat.
  !!
  subroutine fd4_stat_free(stat)
    type(fd4_stat), intent(inout) :: stat

    if(associated(stat%values))     deallocate(stat%values)
    if(associated(stat%names))      deallocate(stat%names)
    if(associated(stat%dumpifset))  deallocate(stat%dumpifset)
    stat%values => null()
    stat%names => null()
    stat%dumpifset => null()
    stat%nstat = 0
    
  end subroutine fd4_stat_free


  !! Count given statistic up by given value.
  !!
  subroutine fd4_stat_countup(stat, idx, value)
    type(fd4_stat), intent(inout) :: stat       !! the statistics object
    integer(i_k), intent(in) :: idx             !! index of the statistic
    integer(i8k), intent(in) :: value           !! value to add
    
    stat%values(idx) = stat%values(idx) + value
    
  end subroutine fd4_stat_countup
  

  !! Collect global min/max/avg/sum values of the FD4 statistics and let rank $opt_root
  !! print the statistics to unit $out.
  !! The default unit is stdout. The default root is rank 0.
  !!
  subroutine fd4_stat_dump(stat, mpi, opt_root, opt_out)
    type(fd4_stat), intent(in)   :: stat             !! the statistics object
    type(fd4_mpi), intent(inout) :: mpi              !! MPI context
    integer, intent(in), optional :: opt_root        !! root MPI process that prints statistics
    integer, intent(in), optional :: opt_out         !! unit for output (use $FD4_STAT_STDOUT for stdout)
    
    integer, parameter :: S_MIN=1, S_MAX=2, S_SUM=3
    integer :: err, i, out, root
    integer(i8k), allocatable :: collect(:,:)
#ifdef FD4_STAT_DUMP_SFCBND
    integer(i_k) :: sfcbnd(2)
    integer, parameter :: ofile=8989
    integer, save :: ocnt=1
#endif
    
    out = fd4_out
    if(present(opt_out)) out = opt_out

    root = 0
    if(present(opt_root)) root = opt_root
    
    allocate(collect(stat%nstat,3))
    
    call MPI_Reduce(stat%values(1), collect(1,S_MIN), stat%nstat, MPI_I8K, MPI_MIN, root, mpi%comm, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Reduce failed')
    call MPI_Reduce(stat%values(1), collect(1,S_MAX), stat%nstat, MPI_I8K, MPI_MAX, root, mpi%comm, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Reduce failed')
    call MPI_Reduce(stat%values(1), collect(1,S_SUM), stat%nstat, MPI_I8K, MPI_SUM, root, mpi%comm, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Reduce failed')
    
    if(mpi%rank==root) then
      write (out,'(A,A20,A13,A13,A13,A19)') FD4_LINEHD,'FD4 Statistics      ','min','max','avg','sum'
      do i=1,stat%nstat
        if(.not. stat%dumpifset(i) .or. collect(i,S_SUM)>0.0_i8k) then
          write (out,'(A,A20,I13,I13,I13,I19)') FD4_LINEHD,stat%names(i), & 
                 collect(i,S_MIN), collect(i,S_MAX), collect(i,S_SUM)/mpi%nprocs, collect(i,S_SUM)
        end if
      end do
    end if
    
    deallocate(collect)

#ifdef FD4_STAT_DUMP_SFCBND
    !! dump min and max SFC index of each rank when SFC partitioning has been used
    if(stat%sfcbnd(2)>0) then
      if(mpi%rank==root) then
        open(unit=ofile,file='fd4stat-sfcbnd-'//i2s(ocnt,.true.)//'.txt',action='write',position='rewind',status='replace')
        write(ofile,'(A,A9,A10,A10)') '#','rank','lbound','ubound'
        do i=0,mpi%nprocs-1
          if(i==root) then
            sfcbnd = stat%sfcbnd
          else
            call MPI_Recv(sfcbnd(1), 2, MPI_I_K, i, 1234, mpi%comm, MPI_STATUS_IGNORE, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Recv failed')
          end if
          write(ofile,'(I10,I10,I10)') i, sfcbnd(1:2)
        end do
        close(ofile)
      else
        call MPI_Send(stat%sfcbnd(1), 2, MPI_I_K, root, 1234, mpi%comm, err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Send failed')
      end if
    end if
#endif
  
  end subroutine fd4_stat_dump

end module fd4_stat_mod
