! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2013 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                       3D Hilbert Curve                       !
!                                                              !
!--------------------------------------------------------------!

#ifdef SFC_ERRCHECK
#define PURE
#endif

#ifndef SFC_MAX_LEV
#define SFC_MAX_LEV 12
#endif

!! Simple 3D Hilbert curve subroutines.
!! Based on a table description of the curve. Parts of the table came from http://wiki.tcl.tk/8723.
!!
!! Routines accept position and hilbert-index starting at 1 (instead of 0 in hilbert.F90)
!!
!! @author Matthias Lieber
module hilbert_mod

  use kinds_mod
  implicit none
  private  
  
  public :: hilbert_i2c, hilbert_c2i
  public :: hilbert_i2c_qsz, hilbert_c2i_qsz
  public :: hilbert_iter, hilbert_iter_init, hilbert_iter_set, hilbert_iter_next
  
  !! Type for one entry in the 3D Hilbert curve description table.
  !!
  !! Values of the $mv item are:
  !! - 1/-1 means x+1/x-1 
  !! - 2/-2 means y+1/y-1 
  !! - 3/-3 means z+1/z-1
  type hilbert_table
    integer(i_k) :: mv(0:6)      !! spatial moves of the curve
    integer(i_k) :: next(0:7)    !! next states when refinining (octant is sequence number)
    integer(i_k) :: num2ref(0:7) !! sequence number on curve primitve to reference (spatial) octant 
    integer(i_k) :: ref2num(0:7) !! reference (spatial) octant to sequence number on curve primitve
  end type hilbert_table  
  
  integer(i_k), parameter :: SFC_DIMS = 3      !! spatial dimensions of SFC
  
  ! (parts of the) table from http://wiki.tcl.tk/8723
  type(hilbert_table), parameter :: tab(0:23) =                                                                       &
  (/                                                                                                                  &
    hilbert_table( (/ 1, 3,-1, 2, 1,-3,-1/), (/ 1, 2, 2,20,20, 4, 4, 5/), (/0,1,5,4,6,7,3,2/), (/0,1,7,6,3,2,4,5/) ), &
    hilbert_table( (/ 3, 2,-3, 1, 3,-2,-3/), (/ 6, 7, 7,19,19, 9, 9,10/), (/0,4,6,2,3,7,5,1/), (/0,7,3,4,1,6,2,5/) ), &
    hilbert_table( (/ 1, 2,-1, 3, 1,-2,-1/), (/11, 0, 0,23,23,14,14,15/), (/0,1,3,2,6,7,5,4/), (/0,1,3,2,7,6,4,5/) ), &
    hilbert_table( (/-1,-3, 1, 2,-1, 3, 1/), (/16,17,17, 7, 7,13,13, 8/), (/5,4,0,1,3,2,6,7/), (/2,3,5,4,1,0,6,7/) ), &
    hilbert_table( (/ 1,-2,-1,-3, 1, 2,-1/), (/19,14,14,10,10, 0, 0,22/), (/6,7,5,4,0,1,3,2/), (/4,5,7,6,3,2,0,1/) ), &
    hilbert_table( (/ 3,-2,-3,-1, 3, 2,-3/), (/23, 9, 9,15,15, 7, 7,21/), (/3,7,5,1,0,4,6,2/), (/4,3,7,0,5,2,6,1/) ), &
    hilbert_table( (/ 2, 1,-2, 3, 2,-1,-2/), (/ 0,11,11,13,13,15,15,14/), (/0,2,3,1,5,7,6,4/), (/0,3,1,2,7,4,6,5/) ), &
    hilbert_table( (/ 3, 1,-3, 2, 3,-1,-3/), (/ 2, 1, 1, 3, 3, 5, 5, 4/), (/0,4,5,1,3,7,6,2/), (/0,3,7,4,1,2,6,5/) ), &
    hilbert_table( (/-3,-2, 3, 1,-3, 2, 3/), (/21,18,18,11,11,20,20,23/), (/6,2,0,4,5,1,3,7/), (/2,5,1,6,3,4,0,7/) ), &
    hilbert_table( (/ 3,-1,-3,-2, 3, 1,-3/), (/13, 5, 5,14,14, 1, 1,17/), (/3,7,6,2,0,4,5,1/), (/4,7,3,0,5,6,2,1/) ), &
    hilbert_table( (/ 2,-1,-2,-3, 2, 1,-2/), (/ 3,15,15, 4, 4,11,11,12/), (/5,7,6,4,0,2,3,1/), (/4,7,5,6,3,0,2,1/) ), &
    hilbert_table( (/ 2, 3,-2, 1, 2,-3,-2/), (/ 7, 6, 6, 8, 8,10,10, 9/), (/0,2,6,4,5,7,3,1/), (/0,7,1,6,3,4,2,5/) ), &
    hilbert_table( (/-1, 3, 1,-2,-1,-3, 1/), (/ 5,13,13,18,18,17,17, 1/), (/3,2,6,7,5,4,0,1/), (/6,7,1,0,5,4,2,3/) ), &
    hilbert_table( (/-1,-2, 1, 3,-1, 2, 1/), (/22,12,12, 6, 6, 3, 3,19/), (/3,2,0,1,5,4,6,7/), (/2,3,1,0,5,4,6,7/) ), &
    hilbert_table( (/ 1,-3,-1,-2, 1, 3,-1/), (/ 8, 4, 4, 9, 9, 2, 2,16/), (/6,7,3,2,0,1,5,4/), (/4,5,3,2,7,6,0,1/) ), &
    hilbert_table( (/ 2,-3,-2,-1, 2, 3,-2/), (/20,10,10, 5, 5, 6, 6,18/), (/5,7,3,1,0,2,6,4/), (/4,3,5,2,7,0,6,1/) ), &
    hilbert_table( (/-3, 2, 3,-1,-3,-2, 3/), (/10,20,20,22,22,18,18, 6/), (/5,1,3,7,6,2,0,4/), (/6,1,5,2,7,0,4,3/) ), &
    hilbert_table( (/-1, 2, 1,-3,-1,-2, 1/), (/15, 3, 3,21,21,12,12,11/), (/5,4,6,7,3,2,0,1/), (/6,7,5,4,1,0,2,3/) ), &
    hilbert_table( (/-3, 1, 3,-2,-3,-1, 3/), (/ 4, 8, 8,12,12,16,16, 2/), (/6,2,3,7,5,1,0,4/), (/6,5,1,2,7,4,0,3/) ), &
    hilbert_table( (/-2,-3, 2, 1,-2, 3, 2/), (/18,21,21, 1, 1,23,23,20/), (/6,4,0,2,3,1,5,7/), (/2,5,3,4,1,6,0,7/) ), &
    hilbert_table( (/-3,-1, 3, 2,-3, 1, 3/), (/17,16,16, 0, 0, 8, 8,13/), (/5,1,0,4,6,2,3,7/), (/2,1,5,6,3,0,4,7/) ), &
    hilbert_table( (/-2, 1, 2,-3,-2,-1, 2/), (/14,19,19,17,17,22,22, 0/), (/6,4,5,7,3,1,0,2/), (/6,5,7,4,1,2,0,3/) ), &
    hilbert_table( (/-2, 3, 2,-1,-2,-3, 2/), (/ 9,23,23,16,16,21,21, 7/), (/3,1,5,7,6,4,0,2/), (/6,1,7,0,5,2,4,3/) ), &
    hilbert_table( (/-2,-1, 2, 3,-2, 1, 2/), (/12,22,22, 2, 2,19,19, 3/), (/3,1,0,2,6,4,5,7/), (/2,1,3,0,5,6,4,7/) )  &
   /)
  
  !! Iterator to traverse the 3D Hilbert curve
  !!
  type hilbert_iter
    integer(i_k) :: idx = - 1
    integer(i_k) :: p(SFC_DIMS) = - 1
    integer(i_k) :: lev = -1
    integer(i_k) :: qsz(SFC_DIMS) = -1
    integer(i_k) :: len = -1
    integer(i_k) :: clev = -1
    integer(i_k) :: s(SFC_MAX_LEV) = -1
    integer(i_k) :: c(SFC_MAX_LEV) = -1
    integer(i_K) :: jump_level = SFC_MAX_LEV + 1
  end type hilbert_iter

contains

include 'util/math.incf'

!! Get index on curve at given 3D coordinate (coordinates-to-index).
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
PURE subroutine hilbert_c2i(lev, p, idx)

  integer(i_k), intent(in)  :: lev    !! level of the hilbert curve (1=2x2x2, 2=4x4x4, 3=8x8x8, etc)
  integer(i_k), intent(in)  :: p(3)   !! 3D coordinates [1 ... 2^$lev]
  integer(i_k), intent(out) :: idx    !! index on the 3D hilbert curve [1 ... (2^$lev)^3]
  
  integer(i_k) :: l,len,dim,q,s,rq,cp(3)
    
  cp(:)=p(:)-1
  idx=1
  s=0
  dim=2**lev
  ! length of curve in base octant
  len=dim**3

#ifdef SFC_ERRCHECK  
  ! very expensive check
  if(cp(1)>=dim .or. cp(1)<0 .or. cp(2)>=dim .or. cp(2)<0 .or. cp(3)>=dim .or. cp(3)<0) then
    print *,'hilbert_c2i: position out of bounds', cp
    stop
  end if
#endif

  do l=lev,1,-1

    ! dimension and length of refined octants
    dim=dim/2
    len=len/8
    
    ! determine spatial octant and reduce spatial coordinates
    ! so that they are relative to refined octant
    ! reference octant:
    ! z=0 | z=1  
    ! 2 3 | 6 7
    ! 0 1 | 4 5
    rq=0
    if(cp(1)>=dim) then
      rq=ior(1,rq)
      cp(1)=cp(1)-dim
    end if
    if(cp(2)>=dim) then
      rq=ior(2,rq)
      cp(2)=cp(2)-dim
    end if
    if(cp(3)>=dim) then
      rq=ior(4,rq)
      cp(3)=cp(3)-dim
    end if
    
    ! get sequence number of octant q in the curve primitive
    q=tab(s)%ref2num(rq)
    
    ! add offset to index    
    idx=idx+q*len
    
    ! get next state
    s=tab(s)%next(q)
    
  end do
  
end subroutine hilbert_c2i


!! Get 3D coordinate at given index on curve (index-to-coordinates).
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
PURE subroutine hilbert_i2c(lev, idx, p)

  integer(i_k), intent(in)  :: lev   !! level of the hilbert curve (1=2x2x2, 2=4x4x4, 3=8x8x8, etc)
  integer(i_k), intent(in)  :: idx   !! index on the 3D hilbert curve [1 ... (2^$lev)^3 ]
  integer(i_k), intent(out) :: p(3)  !! 3D coordinates [1 ... 2^$lev]

  integer(i_k) :: l,len,dim,q,cidx,s,rq
  
  p(:)=1
  s=0
  cidx=idx-1
  dim=2**lev
  
  ! length of curve in base octant
  len=dim**3

#ifdef SFC_ERRCHECK  
  if(cidx>=len .or. cidx<0) then
    print *,'hilbert_i2c: cidx >= len or cidx < 0: ', cidx
    stop
  end if
#endif

  do l=lev,1,-1

    ! dimension of refined octants
    dim=dim/2
    
    ! determine sequence number of octant in the curve primitive.
    ! this alorithm is faster than doing simply this:
    !   len=len/8 ;  q=cidx/len
    ! or
    !   q=(8*cidx)/len
    if(cidx<len/2) then
      if(cidx<len/4) then
        if(cidx<len/8) then
          q=0
        else
          q=1
        end if
      else
        if(cidx<3*len/8) then
          q=2
        else
          q=3
        end if
      end if
    else
      if(cidx<3*len/4) then
        if(cidx<5*len/8) then
          q=4
        else
          q=5
        end if
      else
        if(cidx<7*len/8) then
          q=6
        else
          q=7
        end if
      end if
    end if
    
    ! get actual spatial octant, depending on curve primitive
    ! reference octant:
    ! z=0 | z=1  
    ! 2 3 | 6 7
    ! 0 1 | 4 5
    rq=tab(s)%num2ref(q)
    
    ! update coordinates
    if(iand(1,rq)==1) p(1)=p(1)+dim
    if(iand(2,rq)==2) p(2)=p(2)+dim
    if(iand(4,rq)==4) p(3)=p(3)+dim
    
    ! length of curve in refined octant
    len=len/8
    
    ! decrease idx by offset in current octant
    cidx=cidx-q*len
    
    ! next state
    s=tab(s)%next(q)
    
  end do
 
end subroutine hilbert_i2c






!! Get index on curve at given 3D coordinate (coordinates-to-index).
!!
!! The size of the query region is restricted by $qsz for each dimension.
!! $qsz must be <= $lev**2 and $pos must be <= $qsz.
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
PURE subroutine hilbert_c2i_qsz(lev, qsz, p, idx)

  integer(i_k), intent(in)  :: lev    !! level of the hilbert curve (1=2x2x2, 2=4x4x4, 3=8x8x8, etc)
  integer(i_k), intent(in)  :: qsz(3) !! query region size
  integer(i_k), intent(in)  :: p(3)   !! 3D coordinates [1 ... 2^$lev]
  integer(i_k), intent(out) :: idx    !! index on the 3D hilbert curve [1 ... (2^$lev)^3]
  
  integer(i_k) :: l,dim,q,s,rq,cp(3)
  integer(i_k) :: csz(3), iq, irq, rsz(3)
    
  cp(:)=p(:)-1
  idx=1
  s=0
  dim=2**lev

  csz = qsz

#ifdef SFC_ERRCHECK  
  ! very expensive check
  if(cp(1)>=qsz(1) .or. cp(1)<0 .or. cp(2)>=qsz(2) .or. cp(2)<0 .or. cp(3)>=qsz(3) .or. cp(3)<0) then
    print *,'hilbert_c2i: position out of bounds', cp
    stop
  end if
#endif

  do l=lev,1,-1

    ! dimension and length of refined octants
    dim=dim/2
    
    ! determine spatial octant and reduce spatial coordinates
    ! so that they are relative to refined octant
    ! reference octant:
    ! z=0 | z=1  
    ! 2 3 | 6 7
    ! 0 1 | 4 5
    rq=0
    if(cp(1)>=dim) then
      rq=ior(1,rq)
      cp(1)=cp(1)-dim
    end if
    if(cp(2)>=dim) then
      rq=ior(2,rq)
      cp(2)=cp(2)-dim
    end if
    if(cp(3)>=dim) then
      rq=ior(4,rq)
      cp(3)=cp(3)-dim
    end if
   
    ! get sequence number of octant q in the curve primitive
    q=tab(s)%ref2num(rq)
    
    ! add offset to index
    do iq=0,q-1
      ! count number of cells per quadrant that come before the current quadrant q
      irq = tab(s)%num2ref(iq)
      if(iand(irq,1)==0) then
        rsz(1) = min(dim, csz(1))
      else
        rsz(1) = max(0  , csz(1)-dim)
      end if
      if(iand(irq,2)==0) then
        rsz(2) = min(dim, csz(2))
      else
        rsz(2) = max(0  , csz(2)-dim)
      end if
      if(iand(irq,4)==0) then
        rsz(3) = min(dim, csz(3))
      else
        rsz(3) = max(0  , csz(3)-dim)
      end if
      idx = idx + PRODUCT(rsz)
    end do
    
    ! shrink the current query size (csz) relative to current quadrant
    if(iand(rq,1)>0) then
      csz(1) = csz(1) - dim
    else
      csz(1) = min(csz(1), dim)
    end if
    if(iand(rq,2)>0) then
      csz(2) = csz(2) - dim
    else
      csz(2) = min(csz(2), dim)
    end if
    if(iand(rq,4)>0) then
      csz(3) = csz(3) - dim
    else
      csz(3) = min(csz(3), dim)
    end if
    
    ! get next state
    s=tab(s)%next(q)
    
  end do
  
end subroutine hilbert_c2i_qsz


!! Get 3D coordinate at given index on curve (index-to-coordinates).
!!
!! The size of the query region is restricted by $qsz for each dimension.
!! $qsz must be <= $lev**2 and $pos must be <= $qsz.
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
PURE subroutine hilbert_i2c_qsz(lev, qsz, idx, p, iter)

  integer(i_k), intent(in)  :: lev    !! level of the hilbert curve (1=2x2x2, 2=4x4x4, 3=8x8x8, etc)
  integer(i_k), intent(in)  :: qsz(3) !! query region size
  integer(i_k), intent(in)  :: idx    !! index on the 3D hilbert curve [1 ... (2^$lev)^3 ]
  integer(i_k), intent(out) :: p(3)   !! 3D coordinates [1 ... 2^$lev]
  type(hilbert_iter), optional, intent(inout) :: iter !! for internal use only

  integer(i_k) :: l,dim,q,cidx,s,rq
  integer(i_k) :: csz(3), iq, irq, rsz(3), newcidx
  
  p(:)=1
  s=0
  cidx=idx-1
  dim=2**lev
  
  csz = qsz

#ifdef SFC_ERRCHECK  
  if(cidx>=PRODUCT(qsz) .or. cidx<0) then
    print *,'hilbert_i2c: cidx >= PRODUCT(qsz) or cidx < 0: ', cidx
    stop
  end if
#endif

  do l=lev,1,-1

    ! dimension of refined octants
    dim=dim/2
    
    ! decrease cidx by offset in current quadrant
    ! and determine sequence number q of octant in the curve primitive.
    q=7
    do iq=0,6
      ! count number of cells per quadrant that come before the current quadrant q
      irq=tab(s)%num2ref(iq)
      if(iand(irq,1)==0) then
        rsz(1) = min(dim, csz(1))
      else
        rsz(1) = max(0  , csz(1)-dim)
      end if
      if(iand(irq,2)==0) then
        rsz(2) = min(dim, csz(2))
      else
        rsz(2) = max(0  , csz(2)-dim)
      end if
      if(iand(irq,4)==0) then
        rsz(3) = min(dim, csz(3))
      else
        rsz(3) = max(0  , csz(3)-dim)
      end if
      newcidx = cidx - PRODUCT(rsz)
      if(newcidx>=0) then
        cidx = newcidx
      else
        q=iq
        exit
      end if
    end do

    ! get actual spatial octant, depending on curve primitive
    ! reference octant:
    ! z=0 | z=1  
    ! 2 3 | 6 7
    ! 0 1 | 4 5
    rq=tab(s)%num2ref(q)
    
    ! shrink the current query size (csz) relative to current quadrant
    ! and update coordinates
    if(iand(rq,1)>0) then
      p(1)=p(1)+dim
      csz(1) = csz(1) - dim
    else
      csz(1) = min(csz(1), dim)
    end if
    if(iand(rq,2)>0) then
      p(2)=p(2)+dim
      csz(2) = csz(2) - dim
    else
      csz(2) = min(csz(2), dim)
    end if
    if(iand(rq,4)>0) then
      csz(3) = csz(3) - dim
      p(3)=p(3)+dim
    else
      csz(3) = min(csz(3), dim)
    end if

    if(present(iter)) then
      iter%s(l)=s
      iter%c(l)=q
    end if

    ! next state
    s=tab(s)%next(q)
    
  end do

  if(present(iter)) then
    iter%clev = 1
    do while(iter%c(iter%clev)>6)
      ! go up one level
      iter%clev = iter%clev + 1
      !print *,' next level+',iter%clev
    end do
  end if 
 
end subroutine hilbert_i2c_qsz


!! Initialize the iterator.
!!
!! The current index is $iter%$idx and the current 3D position $iter%$p.
!!
!!> call hilbert_iter_init(iter, lev, qsz)
!!  do while(iter%idx > 0)
!!    array(iter%p(1),iter%p(2),iter%p(3)) = iter%idx
!!    call hilbert_iter_next(iter)
!!  end do
!!< 
!!
PURE subroutine hilbert_iter_init(iter, lev, opt_qsz)
  type(hilbert_iter), intent(inout) :: iter
  integer(i_k), intent(in) :: lev
  integer(i_k), intent(in), optional :: opt_qsz(SFC_DIMS)
  
  integer :: l
  
  if(lev>SFC_MAX_LEV) then
    iter%idx = -1
#ifdef SFC_ERRCHECK  
    print *,'hilbert_iter_init: lev > SFC_MAX_LEV = ',SFC_MAX_LEV
#endif
    return
  end if
  
  iter%idx = 1
  iter%p = 1
  iter%c = 0
  iter%lev = lev
  if(present(opt_qsz)) then
    iter%qsz = opt_qsz
    ! define the level at which we start jumping depending on the smallest dimension of the query size
    iter%jump_level = MAX( log2i(MINVAL(iter%qsz)) - 2, 1)
  else
    iter%qsz = 2**lev
    iter%jump_level = SFC_MAX_LEV + 1
  end if
  iter%len = PRODUCT(iter%qsz)
  
  iter%clev = 1
  iter%s(iter%lev) = 0
  do l=iter%lev,2,-1
    iter%s(l-1) = tab(iter%s(l))%next(0)
  end do
  
end subroutine hilbert_iter_init


!! Set the iterator to given index
!!
PURE subroutine hilbert_iter_set(iter, idx)
  type(hilbert_iter), intent(inout) :: iter
  integer(i_k), intent(in) :: idx
  
  iter%idx = idx
  call hilbert_i2c_qsz(iter%lev, iter%qsz, iter%idx, iter%p, iter)
  
end subroutine hilbert_iter_set


!! Iterate to next point on the hilbert curve
!!
!! After reaching the end of the curve, $iter%$idx is set to 0.
!!
PURE subroutine hilbert_iter_next(iter)
  type(hilbert_iter), intent(inout) :: iter
  
  integer(i_k) :: s, c, l
  logical :: outside_qsz
  
  iter%idx = iter%idx + 1
  
  do

    if(iter%idx>iter%len .or. iter%clev>iter%lev) then
      ! end of iteration
      iter%idx = 0
      exit
    end if

    s=iter%s(iter%clev)
    c=iter%c(iter%clev)
    iter%p(abs(tab(s)%mv(c))) = iter%p(abs(tab(s)%mv(c))) + sign(1,tab(s)%mv(c)) 
    !print *,iter%p,' |',iter%idx,iter%clev,c,s
    
    outside_qsz = ANY(iter%p>iter%qsz)

    if(outside_qsz .and. iter%clev>iter%jump_level) then
    !if(.false.) then

      ! jump if new point out of query region
      call hilbert_i2c_qsz(iter%lev, iter%qsz, iter%idx, iter%p, iter)
      !print *,' jumping to ', iter%idx,' at',iter%p
      !print *,'      *c*',iter%c(1:iter%lev)
      !print *,'      *s*',iter%s(1:iter%lev)
      outside_qsz = .false.

    else ! not jumping: iterate to next point

      iter%c(iter%clev) = iter%c(iter%clev) + 1
      if(iter%clev>1) then
        do while(iter%clev>1)
          ! go down one level
          l = iter%clev - 1
          !print *,' next level-',l,iter%s(iter%clev),iter%c(iter%clev)
          iter%s(l) = tab(iter%s(iter%clev))%next(iter%c(iter%clev))
          iter%c(l) = 0
          iter%clev = l
        end do
      else
        !iter%c(iter%clev) = iter%c(iter%clev) + 1
        do while(iter%c(iter%clev)>6)
          ! go up one level
          iter%clev = iter%clev + 1
          !print *,' next level+',iter%clev
        end do
      end if

    end if ! not jumping
    
    if(.not.outside_qsz) exit
    !iter%idx = iter%idx - 1 ; exit
    
  end do
  

end subroutine hilbert_iter_next


end module hilbert_mod
