! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2013 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                        Red Black Tree                        !
!                                                              !
!--------------------------------------------------------------!

!  The tree stores pointers to a derived type fd4_mdsd.
!  The real name of the type (TYPE_NAME) has to be inserted in this code by:
!   cat rbtree.F90 | sed -e 's/fd4_mdsd/TYPE_NAME/g' > myrbtree.F90
!  It is assumed that the type definition resides in a module called
!  TYPE_NAME_mod.


!!  Red-Black binary search tree combined with a doubly-linked list for fast iteration.
!!
!!  The compexity of the insert, delete, and get (i.e. find) routines is logarithmic.
!!  Iterator functions next and previous have constant complexity.
!!  The tree stores pointers to a derived type $fd4_mdsd.
!!  It is assumed that the type definition resides in a  module called
!!  $fd4_mdsd_mod.
!!  
!!  The Red-Black tree is coded following the descripton from Julienne Walker at
!!  http://www.eternallyconfuzzled.com.
!!
!!  @author Matthias Lieber
module rbtree_fd4_mdsd_mod

  use fd4_mdsd_mod, only: fd4_mdsd
  use kinds_mod, only: i_k
  implicit none
  private
  
  !! Helper to build an array of pointers to tree nodes. Only used internally.
  type rb_node_fd4_mdsd_links
    type(rb_node_fd4_mdsd), pointer :: n => NULL()
  end type rb_node_fd4_mdsd_links
  
  !! A tree node. Only used internally.
  type rb_node_fd4_mdsd
    type(rb_node_fd4_mdsd_links) :: link(0:1)   !! pointers to left and right children
    type(rb_node_fd4_mdsd_links) :: list(0:1)   !! pointers to previous and next node in sorted list
    integer(i_k) :: red = 0                 !! color of the node, 1 if red, 0 if black
    integer(i_k) :: key = 0                 !! the key of the data
    type(fd4_mdsd), pointer :: value => NULL()  !! pointer to a type containing the actual data
  end type rb_node_fd4_mdsd

  !! The tree.
  type rbtree_fd4_mdsd
    type(rb_node_fd4_mdsd), pointer :: root => NULL()  !! pointer to root node
    type(rb_node_fd4_mdsd_links) :: list(0:1)          !! pointers to first and last node in sorted list
    integer(i_k) :: length = 0                       !! number of elements in the tree
  end type rbtree_fd4_mdsd

  !! A tree iterator type.
  type rbtree_fd4_mdsd_iter
    type(rbtree_fd4_mdsd), pointer :: rbtree                 !! pointer to the tree to iterate
    type(rb_node_fd4_mdsd), pointer :: cur => NULL()         !! pointer to the current node
  end type rbtree_fd4_mdsd_iter


  ! RB Tree stuff
  public :: fd4_mdsd
  public :: rbtree_fd4_mdsd, rb_node_fd4_mdsd, rb_node_fd4_mdsd_links
  public :: rbtree_fd4_mdsd_insert, rbtree_fd4_mdsd_delete
  public :: rbtree_fd4_mdsd_validate, rbtree_fd4_mdsd_get
  public :: rbtree_fd4_mdsd_print, rbtree_fd4_mdsd_clear, rbtree_fd4_mdsd_detach, rbtree_fd4_mdsd_swap

  ! RB Tree Iterator stuff
  public :: rbtree_fd4_mdsd_iter
  public :: rbtree_fd4_mdsd_iter_first, rbtree_fd4_mdsd_iter_next
  public :: rbtree_fd4_mdsd_iter_last, rbtree_fd4_mdsd_iter_prev
  public :: rbtree_fd4_mdsd_iter_key, rbtree_fd4_mdsd_iter_value
  public :: rbtree_fd4_mdsd_iter_set


contains

!--------------------------------------------------------------!
!                                                              !
!                    INTERNAL SUBROUTINES                      !
!                                                              !
!--------------------------------------------------------------!

  !! Check if node is red. Internal.
  function is_red (node)
    logical :: is_red
    type(rb_node_fd4_mdsd), pointer :: node
    if(associated(node)) then
      is_red = (node%red==1)
    else
      is_red = .false.
    end if
  end function is_red
  
  !! Single rotation. Internal.
  function single_rotate (node, dir)
    type(rb_node_fd4_mdsd), pointer :: single_rotate  !! return value, new root node of the subtree
    type(rb_node_fd4_mdsd), pointer :: node           !! pivot node
    integer(i_k), intent(in) :: dir               !! direction of rotation, 0 means left, 1 means right
    single_rotate => node%link(1-dir)%n
    node%link(1-dir)%n => single_rotate%link(dir)%n
    single_rotate%link(dir)%n => node
    node%red = 1
    single_rotate%red = 0
  end function single_rotate

  ! Double rotation. Internal.
  function double_rotate (node, dir)
    type(rb_node_fd4_mdsd), pointer :: double_rotate  !! return value, new root node of the subtree
    type(rb_node_fd4_mdsd), pointer :: node           !! pivot node
    integer(i_k), intent(in) :: dir               !! direction of rotation, 0 means left, 1 means right
    node%link(1-dir)%n => single_rotate( node%link(1-dir)%n, 1-dir)
    double_rotate => single_rotate( node, dir)
  end function double_rotate
  
  !! Check if tree is a valid red black tree. Internal.
  recursive function validate (node) RESULT(height)
    integer(i_k) :: height
    type(rb_node_fd4_mdsd), pointer :: node
    integer(i_k) :: lh, rh
    type(rb_node_fd4_mdsd), pointer :: ln, rn
    if(.not. associated(node)) then
      height = 1
      return    
    end if
    ln => node%link(0)%n
    rn => node%link(1)%n
    ! validate red nodes
    if(is_red(node)) then
      if(is_red(ln) .or. is_red(rn)) then
        write(*,*) 'red violation'
        height = 0
        return
      end if
    end if
    lh = validate(ln)
    rh = validate(rn)
    ! validate binary search tree
    if(associated(ln)) then
      if(ln%key >= node%key) then
        write(*,*) 'binary tree violation left'
        height = 0
        return
      end if
    end if
    if(associated(rn)) then
      if(rn%key <= node%key) then
        write(*,*) 'binary tree violation right'
        height = 0
        return
      end if
    end if
    ! validate black height
    if(lh/=0 .and. rh/=0 .and. lh/=rh) then
      write(*,*) 'black violation'
      height = 0
      return
    end if
    ! count black nodes
    if(lh/=0 .and. rh/=0) then
      height = lh
      if(.not. is_red(node)) height = height + 1
    else
      height = 0
    end if    
  end function validate
  
  !! Print tree. Internal.
  recursive subroutine print_tree (node, level)
    type(rb_node_fd4_mdsd), pointer :: node
    integer(i_k), intent(in) :: level
    integer(i_k) :: i
    character(len=15) :: str
    if(associated(node)) then
       call print_tree (node%link(1)%n, level+1)
       do i=1,level
          write(*,'(A)',ADVANCE='no') '     '
       end do
       if(is_red(node)) then
         write(*,'(A)',ADVANCE='no') 'r-'
       else
         write(*,'(A)',ADVANCE='no') 'b-'
       end if
       write(str,'(I10)') node%key
       str=adjustl(str)
       write(*,'(A)') str(1:len_trim(str))
       call print_tree (node%link(0)%n, level+1)
    end if
  end subroutine print_tree
  
  !! Create new node. Internal.
  function create_node (key, value)
    type(rb_node_fd4_mdsd), pointer :: create_node  !! return value, the new node just created
    integer(i_k), intent(in) :: key             !! the key
    type(fd4_mdsd), pointer :: value                !! pointer to the data the node should hold
    allocate(create_node)
    create_node%link(0)%n => NULL()
    create_node%link(1)%n => NULL()    
    create_node%red = 1
    create_node%key = key
    create_node%value => value
  end function create_node
  
  !! Check if two node pointers are the same node. Internal.
  function node_equal (node1, node2)
    integer(i_k) :: node_equal                !! return value, 1 if equal, 0 otherwise
    type(rb_node_fd4_mdsd), pointer :: node1      !! first node to compare
    type(rb_node_fd4_mdsd), pointer :: node2      !! second node to compare
!#define WEIRD_NODE_EQUAL
#ifdef WEIRD_NODE_EQUAL
    node_equal=0
    if(associated(node1)) then
      if(associated(node2)) then
        if(node1%key==node2%key) node_equal=1
      end if
    else
      if(.not. associated(node2)) node_equal=1
    end if
#else
    if(associated(node1,node2)) then
      node_equal=1
    else
      node_equal=0
    end if    
#endif
  end function node_equal


!--------------------------------------------------------------!
!                                                              !
!                    INTERFACE SUBROUTINES                     !
!                                                              !
!--------------------------------------------------------------!


  !! Check if tree is valid red black tree.
  !! Prints errors to stdout.
  subroutine rbtree_fd4_mdsd_validate(rbtree, ierr)
    type(rbtree_fd4_mdsd), intent(in) :: rbtree           !! the red black tree
    integer(i_k), optional, intent(out) :: ierr       !! error status
    integer(i_k) :: res
    res = validate(rbtree%root)
    if(present(ierr)) ierr = res    
  end subroutine rbtree_fd4_mdsd_validate


  !! Print the tree to stdout. Useful for debugging.
  subroutine rbtree_fd4_mdsd_print(rbtree)
    type(rbtree_fd4_mdsd), intent(in) :: rbtree           !! the red black tree
    call print_tree(rbtree%root,0)
  end subroutine rbtree_fd4_mdsd_print
 
 
  !! Get a value from the tree. Top-down search without recursion.
  function rbtree_fd4_mdsd_get(rbtree, key)
    type(fd4_mdsd), pointer :: rbtree_fd4_mdsd_get    !! return value, pointer to requested data or $null if not found
    type(rbtree_fd4_mdsd), intent(in) :: rbtree   !! the red black tree
    integer(i_k), intent(in) :: key           !! key of the entry to retrieve
    type(rb_node_fd4_mdsd), pointer :: q
    integer(i_k) :: dir
    
    q => rbtree%root
    do 
      if(.not. associated(q)) then
        ! key not found
        rbtree_fd4_mdsd_get => NULL()          
        exit
      end if
      if(q%key<key) then
        dir=1
      elseif(q%key>key) then
        dir=0
      else
        ! found node
        rbtree_fd4_mdsd_get => q%value
        exit
      end if
      ! go down
      q => q%link(dir)%n
    end do
    
  end function rbtree_fd4_mdsd_get


  !! Insert new entry to the tree. Top-down insertion without recursion.
  subroutine rbtree_fd4_mdsd_insert(rbtree, key, value, ierr)
    type(rbtree_fd4_mdsd), intent(inout) :: rbtree     !! the red black tree
    integer(i_k), intent(in) :: key                !! key of the entry to insert
    type(fd4_mdsd), pointer :: value                   !! pointer to value of the entry to insert
    integer(i_k), optional, intent(out) :: ierr    !! optional error status, 0 means ok, 1 means duplicate key

    type(rb_node_fd4_mdsd), target :: head ! fake tree head
    type(rb_node_fd4_mdsd), pointer :: q, p, g, t ! current, parent, grandparent, grandgrandparent
    logical :: inserted  ! true if we inserted new node this iteration
    integer(i_k) :: dir, ldir, dir2
    
    if(present(ierr)) ierr=0
    
    if(.not. associated(rbtree%root)) then
      ! empty tree case
      rbtree%root => create_node(key, value)
      rbtree%root%red = 0
      rbtree%length = rbtree%length + 1
      rbtree%root%list(0)%n => null()
      rbtree%root%list(1)%n => null()
      rbtree%list(0)%n => rbtree%root
      rbtree%list(1)%n => rbtree%root
      return
    end if
    
    ldir = 0
    dir = 0
    t => head
    g => NULL()
#ifdef CAUTIOUS
    p => NULL()
#else
    ! i'm not 100% sure that we can do that safely, but it improves speed a bit
    ! since we always have a parent and need not to check this
    head%red = 0 ! p = t = head is black
    p => t
#endif
    t%link(1)%n => rbtree%root
    q => t%link(1)%n
    inserted = .false.
    
    do
      if(.not. associated(q)) then
        ! insert new node at the bottom
        p%link(dir)%n => create_node(key, value)      
        rbtree%length = rbtree%length + 1
        q => p%link(dir)%n
        inserted = .true.
      elseif( is_red(q%link(0)%n) .and. is_red(q%link(1)%n) ) then
        ! color flip
        q%red = 1
        q%link(0)%n%red = 0
        q%link(1)%n%red = 0
      end if
      
#ifdef CAUTIOUS
      if(q%red==1 .and. is_red(p)) then
#else
      if(q%red==1 .and. p%red==1) then
#endif
        ! rotation to fix red violation 
        dir2 = node_equal(t%link(1)%n, g)
        if( associated(q,p%link(ldir)%n) ) then
          t%link(dir2)%n => single_rotate(g, 1-ldir)
        else
          t%link(dir2)%n => double_rotate(g, 1-ldir)
        end if      
      end if
      
      ! stop iteration if new node inserted
      if(inserted) exit
      
      ! get new direction
      ldir = dir
      if(q%key<key) then
        dir=1
      elseif(q%key>key) then
        dir=0
      else
        ! duplicate key not allowed
        if(present(ierr)) ierr=1
        !print *,'duplicate key!'
        exit
      end if
      
      ! update parent, grandparent, etc.
      if(associated(g)) t => g
      g => p
      p => q
      q => q%link(dir)%n

    end do
    
    rbtree%root => head%link(1)%n
    rbtree%root%red = 0    

    ! update the list pointers
    if(inserted) then
      ! In fact, p may not be the parent of q by now, due to rotation
      ! after insertion of q. But q has been inserted as child of p
      ! and thus p is the next or previous node of q.
      if(p%key < q%key) then
        ! new node is parent's right child
        q%list(1)%n => p%list(1)%n          ! my next is parent's next                   
        q%list(0)%n => p                    ! my prev is parent                          
        p%list(1)%n => q                    ! my prev's next is me                       
        if(associated(q%list(1)%n)) then
          q%list(1)%n%list(0)%n => q        ! my next's prev is me                       
        else
          rbtree%list(1)%n => q             ! i am the last                              
        end if
      else
        ! new node is parent's left child
        q%list(0)%n => p%list(0)%n          ! my prev is parent's prev                   
        q%list(1)%n => p                    ! my next is parent                          
        p%list(0)%n => q                    ! my next's prev is me                       
        if(associated(q%list(0)%n)) then
          q%list(0)%n%list(1)%n => q        ! my prev's next is me                       
        else
          rbtree%list(0)%n => q             ! i am the first                             
        end if
      end if
    end if

  end subroutine rbtree_fd4_mdsd_insert
 

  !! Delete entry from the tree. Top-down deletion without recursion.
  !!
  !! If an optional iterator is given, it will be repaired when damaged due
  !! deletion. This damage may happen to any iterator of the tree,
  !! so you must do this when deleting entries within a tree iteration!
  !! This damage is due to the delete algorithm, which sometimes swap tree node contents.
  !! In this case, the iterator would point to the wrong node.
  !! *CAUTION*: Do not delete the current entry of the iterator!
  !! If you still do so, you *must* re-initialize the iterator.
  !! The following code snipped shows how to safely delete entries within an iteration:
  !!
  !!>      value => rbtree_fd4_mdsd_iter_first(rbiter, rbtree)
  !!       do while ( associated(value) )
  !!         ! get key of current
  !!         key = rbtree_fd4_mdsd_iter_key(rbiter)
  !!         ! this function returns true if we do not longer need this item
  !!         useless = is_item_useless(value)
  !!         ! iterate one item forward
  !!         value => rbtree_fd4_mdsd_iter_next(rbiter)
  !!         if(useless) then
  !!           ! delete previous item and fix iterator if necessary
  !!           deleted => rbtree_fd4_mdsd_delete(rbtree, key, rbiter)
  !!         end if
  !!<      end do
  !!
  function rbtree_fd4_mdsd_delete(rbtree, key, iter)
    !! return value, pointer to the value deleted from list or $null if key not found
    type(fd4_mdsd), pointer :: rbtree_fd4_mdsd_delete 
    !! the red black tree 
    type(rbtree_fd4_mdsd), intent(inout) :: rbtree
    !! key of the entry to delete
    integer(i_k), intent(in) :: key
    !! an optional iterator which will be repaired if damaged due to deletion
    type(rbtree_fd4_mdsd_iter), optional, intent(inout) :: iter

    type(rb_node_fd4_mdsd), target :: head        ! fake tree head
    type(rb_node_fd4_mdsd), pointer :: q, p, g, s ! current, parent, grandparent, sibling
    type(rb_node_fd4_mdsd), pointer :: f          ! found item to delete
    integer(i_k) :: dir, ldir, dir2
    
    dir = 1
    q => head
    q%link(dir)%n => rbtree%root
    g => NULL()
    p => NULL()
    f => NULL()
    
    do while(associated(q%link(dir)%n))
      
      ! update parent, grandparent, etc.
      g => p
      p => q
      q => q%link(dir)%n

      ! get new direction
      ldir = dir
      if(q%key<key) then
        dir=1
      elseif(q%key>key) then
        dir=0
      else
        ! found node
        f => q
        dir = 0
      end if
    
      if( q%red==1 .or. is_red(q%link(dir)%n) ) cycle
      ! q and q's child (in our traverse direction) are black

      if( is_red(q%link(1-dir)%n) ) then
        ! q's other child is red, perform rotation
        p%link(ldir)%n => single_rotate(q, dir) 
        p => p%link(ldir)%n
      else
        ! q's other child is also black
        s => p%link(1-ldir)%n
        if(.not. associated(s)) cycle
        ! sibling exists (and is always black since parent is red)
        if( is_red(s%link(ldir)%n) ) then
          ! inner child of sibling is red
          dir2 = node_equal(g%link(1)%n, p)
          g%link(dir2)%n => double_rotate(p, ldir) 
          ! recolor
          q%red = 1
          g%link(dir2)%n%red = 1
          g%link(dir2)%n%link(0)%n%red = 0
          g%link(dir2)%n%link(1)%n%red = 0
        else if( is_red(s%link(1-ldir)%n) ) then
          ! outer child of sibling is red
          dir2 = node_equal(g%link(1)%n, p)
          g%link(dir2)%n => single_rotate(p, ldir) 
          ! recolor
          q%red = 1
          g%link(dir2)%n%red = 1
          g%link(dir2)%n%link(0)%n%red = 0
          g%link(dir2)%n%link(1)%n%red = 0
        else
          ! sibling has only black children, do color flip
          p%red=0
          s%red=1
          q%red=1       
        end if
      end if
    end do
    
    if(associated(f)) then
      ! this is the return value
      rbtree_fd4_mdsd_delete => f%value

      ! update the list pointers of q's prev and next nodes
      ! and, if necessary, the first and last pointer of the tree
      do dir=0,1
        if(associated(q%list(dir)%n)) then
          q%list(dir)%n%list(1-dir)%n => q%list(1-dir)%n
        else
          rbtree%list(dir)%n => q%list(1-dir)%n
        end if
      end do

      ! repair iterator
      if(present(iter)) then
        if(associated(iter%cur,q)) iter%cur => f
      end if

      ! move key and value from q to f
      f%key = q%key
      f%value => q%value
      ldir = node_equal(p%link(1)%n,q)  ! direction from p to q
      if(associated(q%link(0)%n)) then
        dir = 0
      else
        dir = 1   ! direction from q to q's child
      end if
      ! let q's child be p's child
      p%link(ldir)%n => q%link(dir)%n
      ! free memory
      deallocate(q)
      rbtree%length = rbtree%length - 1
    else
      rbtree_fd4_mdsd_delete => NULL()          
    end if
    
    rbtree%root => head%link(1)%n
    if(associated(rbtree%root)) rbtree%root%red = 0    

  end function rbtree_fd4_mdsd_delete

  !! Clear the whole tree.
  subroutine rbtree_fd4_mdsd_clear(rbtree, dealloc_sub)
    type(rbtree_fd4_mdsd), intent(inout) :: rbtree                  !! the red black tree
    !! subroutine which deallocates a pointer of type fd4_mdsd
    interface
      subroutine dealloc_sub(p)
        use fd4_mdsd_mod, only: fd4_mdsd
        type(fd4_mdsd), pointer :: p
      end subroutine dealloc_sub
    end interface
    
    type(rb_node_fd4_mdsd), pointer :: cur, next
    
    ! iterate over all nodes
    cur => rbtree%list(0)%n
    do while(associated(cur))
      next => cur%list(1)%n
      ! call subroutine to deallocate data
      call dealloc_sub(cur%value)
      ! deallocate node
      deallocate(cur)    
      cur => next      
    end do
    
    ! set everything to 0
    rbtree%length = 0
    rbtree%root => null()
    rbtree%list(0)%n => null()
    rbtree%list(1)%n => null()    
  
  end subroutine rbtree_fd4_mdsd_clear

  !! Clear the whole tree *without deallocating contained items*.
  subroutine rbtree_fd4_mdsd_detach(rbtree)
    type(rbtree_fd4_mdsd), intent(inout) :: rbtree                  !! the red black tree
    
    type(rb_node_fd4_mdsd), pointer :: cur, next
    
    ! iterate over all nodes
    cur => rbtree%list(0)%n
    do while(associated(cur))
      next => cur%list(1)%n
      ! deallocate node
      deallocate(cur)    
      cur => next      
    end do
    
    ! set everything to 0
    rbtree%length = 0
    rbtree%root => null()
    rbtree%list(0)%n => null()
    rbtree%list(1)%n => null()    
  
  end subroutine rbtree_fd4_mdsd_detach

  !! Swap two rbtrees (just swaps the internal pointers of the tree structure)
  subroutine rbtree_fd4_mdsd_swap(rbtree1, rbtree2)
    type(rbtree_fd4_mdsd), intent(inout) :: rbtree1                  !! the 1st red black tree
    type(rbtree_fd4_mdsd), intent(inout) :: rbtree2                  !! the 2nd red black tree
    
    type(rbtree_fd4_mdsd) :: rbtree_tmp
    
    ! set tmp to 1
    rbtree_tmp%root      => rbtree1%root
    rbtree_tmp%list(0)%n => rbtree1%list(0)%n
    rbtree_tmp%list(1)%n => rbtree1%list(1)%n
    rbtree_tmp%length    =  rbtree1%length

    ! set 1 to 2
    rbtree1%root      => rbtree2%root
    rbtree1%list(0)%n => rbtree2%list(0)%n
    rbtree1%list(1)%n => rbtree2%list(1)%n
    rbtree1%length    =  rbtree2%length
    
    ! set 2 to tmp
    rbtree2%root      => rbtree_tmp%root
    rbtree2%list(0)%n => rbtree_tmp%list(0)%n
    rbtree2%list(1)%n => rbtree_tmp%list(1)%n
    rbtree2%length    =  rbtree_tmp%length
  
  end subroutine rbtree_fd4_mdsd_swap
   
!--------------------------------------------------------------!
!                                                              !
!                     ITERATOR SUBROUTINES                     !
!                                                              !
!--------------------------------------------------------------!


  !! Initialize iterator for tree $rbtree,
  !! position on the first element.
  function rbtree_fd4_mdsd_iter_first(iter, rbtree)
    type(fd4_mdsd), pointer :: rbtree_fd4_mdsd_iter_first   !! return value, pointer to first value in the tree
    type(rbtree_fd4_mdsd_iter), intent(inout) :: iter   !! the iterator
    type(rbtree_fd4_mdsd), target,  intent(in) :: rbtree    !! the red black tree
    iter%cur => rbtree%list(0)%n
    iter%rbtree => rbtree
    if(associated(iter%cur)) then
      rbtree_fd4_mdsd_iter_first => iter%cur%value
    else
      rbtree_fd4_mdsd_iter_first => NULL()
    end if
  end function rbtree_fd4_mdsd_iter_first


  !! Initialize iterator for tree $rbtree,
  !! position on the last element.
  function rbtree_fd4_mdsd_iter_last(iter, rbtree)
    type(fd4_mdsd), pointer :: rbtree_fd4_mdsd_iter_last    !! return value, pointer to last value in the tree
    type(rbtree_fd4_mdsd_iter), intent(inout) :: iter   !! the iterator
    type(rbtree_fd4_mdsd), target,  intent(in) :: rbtree    !! the red black tree
    iter%cur => rbtree%list(1)%n
    iter%rbtree => rbtree
    if(associated(iter%cur)) then
      rbtree_fd4_mdsd_iter_last => iter%cur%value
    else
      rbtree_fd4_mdsd_iter_last => NULL()
    end if
  end function rbtree_fd4_mdsd_iter_last



  !! Iterate to next item in tree.
  !!
  !! *CAUTION*: This routine will intentionally cause a segmentation violation
  !! if the iterator is not valid!
  function rbtree_fd4_mdsd_iter_next(iter)
    type(fd4_mdsd), pointer :: rbtree_fd4_mdsd_iter_next   !! return value, pointer to next value
    type(rbtree_fd4_mdsd_iter), intent(inout) :: iter  !! the iterator

#ifdef RBTREE_DO_CHECKS
    if(.not. associated (iter%cur)) then
      rbtree_fd4_mdsd_iter_next => NULL()
      return
    end if
#endif

    iter%cur => iter%cur%list(1)%n
    
    if(associated(iter%cur)) then
      rbtree_fd4_mdsd_iter_next => iter%cur%value
    else
      rbtree_fd4_mdsd_iter_next => NULL()
    end if

  end function rbtree_fd4_mdsd_iter_next



  !! Iterate to previous item in tree.
  !!
  !! *CAUTION*: This routine will intentionally cause a segmentation violation
  !! if the iterator is not valid!
  function rbtree_fd4_mdsd_iter_prev(iter)
    type(fd4_mdsd), pointer :: rbtree_fd4_mdsd_iter_prev  !! return value, pointer to previous value
    type(rbtree_fd4_mdsd_iter), intent(inout) :: iter !! the iterator

#ifdef RBTREE_DO_CHECKS
    if(.not. associated (iter%cur)) then
      rbtree_fd4_mdsd_iter_prev => NULL()
      return
    end if
#endif

    iter%cur => iter%cur%list(0)%n
    
    if(associated(iter%cur)) then
      rbtree_fd4_mdsd_iter_prev => iter%cur%value
    else
      rbtree_fd4_mdsd_iter_prev => NULL()
    end if

  end function rbtree_fd4_mdsd_iter_prev


  !! Set position of an iterator. Top-down search without recursion.
  function rbtree_fd4_mdsd_iter_set(iter, key)
    !! return value, value at the requested position or $null if not found
    type(fd4_mdsd), pointer :: rbtree_fd4_mdsd_iter_set
    type(rbtree_fd4_mdsd_iter), intent(inout) :: iter  !! the iterator
    integer(i_k), intent(in) :: key                    !! key of the entry to search
    integer(i_k) :: dir

    ! same algorithm as in rbtree_fd4_mdsd_get
    iter%cur => iter%rbtree%root
    do 
      if(.not. associated(iter%cur)) then
        rbtree_fd4_mdsd_iter_set => NULL()
        exit      
      end if
      if(iter%cur%key<key) then
        dir=1
      elseif(iter%cur%key>key) then
        dir=0
      else
        ! found node
        rbtree_fd4_mdsd_iter_set => iter%cur%value
        exit
      end if
      ! go down
      iter%cur => iter%cur%link(dir)%n
    end do
    
  end function rbtree_fd4_mdsd_iter_set


  !! Get the key of current node.
  !!
  !! *CAUTION*: This routine will intionally cause a segmentation violation
  !! if the iterator is not valid!
  function rbtree_fd4_mdsd_iter_key(iter)
    integer(i_k) :: rbtree_fd4_mdsd_iter_key        !! return value, the key of current node
    type(rbtree_fd4_mdsd_iter), intent(in) :: iter  !! the iterator
    rbtree_fd4_mdsd_iter_key = iter%cur%key         ! will intionally cause segfault if iter%cur is null!
  end function rbtree_fd4_mdsd_iter_key



  !! Get value of current node.
  function rbtree_fd4_mdsd_iter_value(iter)
    !! return value, the value of current node or $null if current node is $null
    type(fd4_mdsd), pointer :: rbtree_fd4_mdsd_iter_value
    type(rbtree_fd4_mdsd_iter), intent(in) :: iter      !! the iterator
    if(associated(iter%cur)) then
      rbtree_fd4_mdsd_iter_value => iter%cur%value
    else
      rbtree_fd4_mdsd_iter_value => NULL()  
    end if
  end function rbtree_fd4_mdsd_iter_value


end module rbtree_fd4_mdsd_mod


