! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2013 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!! A section of a 3D domain.
!!
!! The section is made up of a very limited number of clusters each consisting
!! of a 3D rectangle (cuboid).
!!
!! Rectangles can be added to a section, the algorithm tries to reduce the total
!! volume of all clusters.
!!
!! Preprocessor options:
!!  - $OLD_SECTION_ADD: use the old clustering algorithm, which is faster but
!!    has worse quality (much larger clusters)
!!
!! @author Matthias Lieber
module section_mod
  
  use kinds_mod
  implicit none
  
  private

!#define SECTION_ADD_NEW
  
  integer(i_k), parameter :: SECTION_RECT_MAX = 4
  
  !! A 3D section consisting of a limited number of 3D cuboids
  type section
    integer(i_k) :: rect(3,2,SECTION_RECT_MAX) = 0  !! a limited number of 3D cuboids
    integer(i_k) :: nrect = 0                       !! number of used cuboids
    integer(i_k) :: maxext(3)
    integer(i_k) :: vol(SECTION_RECT_MAX) = 0       !! volume of the 3D cuboids
    !! volume of unions of the existing cuboids
    integer(i_k) :: uvol(SECTION_RECT_MAX,SECTION_RECT_MAX) = HUGE(SECTION_RECT_MAX)
  end type section

  public :: section, section_reset, section_add, section_intersect, section_contains, section_volume

contains

  
  !! Reset $sect (set volume to null).
  !!
  !! Parameter $maxrect will be used in $section_add to estimate if it
  !! is useful to add a new cuboid or not.
  !! This argument required only if compiled with $OLD_SECTION_ADD.
  !!
  subroutine section_reset(sect, maxrect)
    type(section), intent(inout) :: sect         !! the section
    integer(i_k), optional, intent(in) :: maxrect(3,2)    !! max allowed size 
    
    sect%nrect = 0
    ! spans a negative section = null
    !sect%rect(1:3,1,1) = maxrect(1:3,2) + 1
    !sect%rect(1:3,2,1) = maxrect(1:3,1) - 1
#ifdef OLD_SECTION_ADD
    ! save maxext
    if(present(maxrect)) then
      sect%maxext = maxrect(1:3,2) - maxrect(1:3,1) + 1
    else
      stop 'you need to pass argument maxrect'
    end if
#endif
    !
    sect%vol(:) = 0
    sect%uvol(:,:) = HUGE(sect%nrect)
    
  end subroutine section_reset


  !! Add $rect to section $sect (union).
  !!
  !! If necessary it will create a new cuboid in $sect to keep overall
  !! volume as small as possible.
  !!
  subroutine section_add(sect, rect)
    type(section), intent(inout) :: sect         !! the section
    integer(i_k),  intent(in) :: rect(3,2)       !! 3D cuboid to add

    integer(i_k) :: v1(SECTION_RECT_MAX), v2(SECTION_RECT_MAX)
    integer(i_k) :: hit, irect
#ifdef OLD_SECTION_ADD
    integer(i_k) :: tmp1(3,2), ext1(3), ext2(3)
#else
    integer(i_k) :: tmp(3,2,SECTION_RECT_MAX), vhit(2), vr
#endif
   
#ifdef OLD_SECTION_ADD

    if(sect%nrect <= 0) then
    
      !! special case: empty $sect
      
      sect%nrect = 1
      sect%rect(1:3,1:2,1) = rect(1:3,1:2)
      
    elseif(sect%nrect < SECTION_RECT_MAX) then
      
      !! we could still create a new rect
      
      hit = 1
      do irect = 1, sect%nrect
        hit = irect
        ! unify sect%rect with rect
        tmp1(1:3,1) = MIN( sect%rect(1:3,1,irect), rect(1:3,1) )
        tmp1(1:3,2) = MAX( sect%rect(1:3,2,irect), rect(1:3,2) )
        ! calculate extents (before/after union)
        ext1(1:3) = sect%rect(1:3,2,irect) - sect%rect(1:3,1,irect) + 1
        ext2(1:3) = tmp1(1:3,2) - tmp1(1:3,1) + 1
        ! relative growth in any dimension > 2 ?
        ! and size of union in this dimension > max size / 4 ?
        if( ANY(ext2 > ext1 * 2 .and. ext2 * 4 > sect%maxext) ) then
            ! this is no good rect for the union, skip this and try next one
            hit = 0
        end if
        ! ok, we found a good rect, let's get out of here
        if(hit /= 0) exit      
      end do


      if(hit /= 0) then
        ! found good rect for union
        sect%rect(1:3,1:2,hit) = tmp1(1:3,1:2)
      else
        ! found no good rect
        ! but we can still create a new one
        sect%nrect = sect%nrect + 1
        sect%rect(1:3,1:2,sect%nrect) = rect(1:3,1:2)
      end if

    else
    
      !! we cannot create a new rect
      !! so just add this rect in such a way, that the sum volume is smallest

      do irect = 1, sect%nrect
        ! unify sect%rect with rect
        tmp1(1:3,1) = MIN( sect%rect(1:3,1,irect), rect(1:3,1) )
        tmp1(1:3,2) = MAX( sect%rect(1:3,2,irect), rect(1:3,2) )
        ! calculate extents (before/after union)
        ext1(1:3) = sect%rect(1:3,2,irect) - sect%rect(1:3,1,irect) + 1
        ext2(1:3) = tmp1(1:3,2) - tmp1(1:3,1) + 1
        ! volumes
        v1(irect) = PRODUCT(ext1)
        v2(irect) = PRODUCT(ext2)        
      end do

      ! calculate in v2 the total volume of all cuboids after adding to sect%rect(irect)
      do irect = 1, sect%nrect
        v2(irect) = v2(irect) + SUM(v1) - v1(irect)
      end do
      
      ! ok, this must be it
      hit = MINLOC(v2,1)
      sect%rect(1:3,1,hit) = MIN( sect%rect(1:3,1,hit), rect(1:3,1) )
      sect%rect(1:3,2,hit) = MAX( sect%rect(1:3,2,hit), rect(1:3,2) )      
      
    end if    
    

#else

      
    vr = PRODUCT(rect(1:3,2) - rect(1:3,1) + 1)
   
    if(sect%nrect <= 0) then
    
      !! special case: empty $sect
      
      sect%nrect = 1
      sect%rect(1:3,1:2,1) = rect(1:3,1:2)
      sect%vol(1) = vr
      
    elseif(sect%nrect < SECTION_RECT_MAX) then
      
      !! we could still create a new rect
      
      do irect = 1, sect%nrect
        ! unify sect%rect with rect
        tmp(1:3,1,irect) = MIN( sect%rect(1:3,1,irect), rect(1:3,1) )
        tmp(1:3,2,irect) = MAX( sect%rect(1:3,2,irect), rect(1:3,2) )
        ! calculate volume after union
        v1(irect) = PRODUCT( tmp(1:3,2,irect) - tmp(1:3,1,irect) + 1 )
        ! volume that is added (rect after union - rect before union - added rect)
        v2(irect) = v1(irect) - sect%vol(irect) - vr
      end do

      hit = MINLOC(v2(1:sect%nrect),1)
      
      !print *,hit,v2(hit),v1(hit)

      if(v2(hit)>0) then
        ! only add to the found rect if <= 1/4 of union are "wasted" cells
        if(4*v2(hit)>v1(hit)) then
          hit=0
        end if
      end if
      
      if(hit /= 0) then
        ! found good rect for union
        sect%rect(1:3,1:2,hit) = tmp(1:3,1:2,hit)
        sect%vol(hit) = v1(hit)
        call section_add_update_uvol(sect, hit)
        ! check if we can unify this rect with any other rect
        call section_add_check_unify(sect, hit)
      else
        ! found no good rect
        ! but we can still create a new one
        sect%nrect = sect%nrect + 1
        sect%rect(1:3,1:2,sect%nrect) = rect(1:3,1:2)
        sect%vol(sect%nrect) = vr
        ! update uvol
        do irect=1,sect%nrect-1
          sect%uvol(sect%nrect,irect) = v2(irect)
        end do
      end if

#ifdef SECTION_ADD_DEBUG
      print *,'+++'
      print *,sect%uvol(1,:)
      do irect = 2, sect%nrect
        print *,sect%uvol(irect,1:irect-1)
      end do
#endif

    else
    
      !! we cannot create a new rect
      !! try to unify with existing rects or try to unify two of the existing rects
      !! and create a new rect

      do irect = 1, sect%nrect
        ! unify sect%rect with rect
        tmp(1:3,1,irect) = MIN( sect%rect(1:3,1,irect), rect(1:3,1) )
        tmp(1:3,2,irect) = MAX( sect%rect(1:3,2,irect), rect(1:3,2) )
        ! calculate volume after union
        v1(irect) = PRODUCT( tmp(1:3,2,irect) - tmp(1:3,1,irect) + 1 )
        ! volume that is added (rect after union - rect before union - added rect)
        v2(irect) = v1(irect) - sect%vol(irect) - vr
        ! set uvol
        sect%uvol(1,irect) = v2(irect)
      end do
      
      vhit = MINLOC(sect%uvol)

#ifdef SECTION_ADD_DEBUG
      print *,'---'
      print *,sect%uvol(1,:)
      do irect = 2, sect%nrect
        print *,sect%uvol(irect,1:irect-1)
      end do
#endif
      
      if(vhit(1)==1) then
        ! add new rect to rect vhit(2)
        sect%rect(:,:,vhit(2)) = tmp(:,:,vhit(2))
        sect%vol(vhit(2)) = v1(vhit(2))
        call section_add_update_uvol(sect, vhit(2))
        ! check if we can unify this rect with any other rect
        call section_add_check_unify(sect, vhit(2))
      else
        ! vhit(1) > vhit(2)
        ! unify rects vhit(1) and vhit(2) in vhit(2)
        sect%rect(1:3,1,vhit(2)) = MIN( sect%rect(1:3,1,vhit(1)), sect%rect(1:3,1,vhit(2)) )
        sect%rect(1:3,2,vhit(2)) = MAX( sect%rect(1:3,2,vhit(1)), sect%rect(1:3,2,vhit(2)) )
        sect%vol(vhit(2)) = PRODUCT( sect%rect(1:3,2,vhit(2)) - sect%rect(1:3,1,vhit(2)) + 1 )
        call section_add_update_uvol(sect, vhit(2))
        ! create new rect in vhit(1)
        sect%rect(1:3,1:2,vhit(1)) = rect(1:3,1:2)
        sect%vol(vhit(1)) = vr
        call section_add_update_uvol(sect, vhit(1))
        ! check if we can unify rect vhit(2) with any other rect
        call section_add_check_unify(sect, vhit(2))
      end if

      
    end if    

#endif        
    
  end subroutine section_add



  !! Update $uvol for cuboid $r.
  !!
  !! Internal routine.
  pure subroutine section_add_update_uvol(sect, r)
    type(section), intent(inout) :: sect         !! the section
    integer(i_k), intent(in) :: r                !! cuboid to update
    integer(i_k) :: irect, tmp(3,2)
    ! update uvol - 2nd dim
    do irect=1,r-1
      tmp(1:3,1) = MIN( sect%rect(1:3,1,irect), sect%rect(1:3,1,r) )
      tmp(1:3,2) = MAX( sect%rect(1:3,2,irect), sect%rect(1:3,2,r) )
      sect%uvol(r,irect) = PRODUCT(tmp(1:3,2) - tmp(1:3,1) + 1) &
                         - sect%vol(r) - sect%vol(irect) 
    end do
    ! update uvol - 1st dim
    do irect=r+1,sect%nrect
      tmp(1:3,1) = MIN( sect%rect(1:3,1,irect), sect%rect(1:3,1,r) )
      tmp(1:3,2) = MAX( sect%rect(1:3,2,irect), sect%rect(1:3,2,r) )
      sect%uvol(irect,r) = PRODUCT(tmp(1:3,2) - tmp(1:3,1) + 1) &
                         - sect%vol(r) - sect%vol(irect) 
    end do
  end subroutine section_add_update_uvol


  !! Unify cuboid $r with other cuboids if this is possible at no cost.
  !!
  !! The processes is repeated for the unified cuboid.
  !!
  !! Internal routine.
  pure subroutine section_add_check_unify(sect, r)
    type(section), intent(inout) :: sect         !! the section
    integer(i_k), intent(in) :: r                !! cuboid to unify
    
    integer(i_k) :: hit, vhit(2)
#ifdef SECTION_ADD_DEBUG
    integer(i_k) :: irect
#endif
    hit = r

    do

#ifdef SECTION_ADD_DEBUG
      print *,'==='
      do irect = 2, sect%nrect
        print *,sect%uvol(irect,1:irect-1)
      end do
#endif

      vhit(1:2)=0
      if(hit>1) then
        if(MINVAL(sect%uvol(hit,1:hit-1))<=0) then
          vhit=MINLOC(sect%uvol(hit:hit,1:hit-1))
          vhit(1)=vhit(1)+hit-1
        end if
      end if
      if(vhit(1)==0 .and. hit<sect%nrect) then
        if(MINVAL(sect%uvol(hit+1:sect%nrect,hit))<=0) then
          vhit=MINLOC(sect%uvol(hit+1:sect%nrect,hit:hit))
          vhit(1)=vhit(1)+hit
          vhit(2)=vhit(2)+hit-1
        end if
      end if
      if(vhit(1)>0) then
        ! found other rect to unify
        ! vhit(1) > vhit(2)
        ! unify rects vhit(1) and vhit(2)
        sect%rect(1:3,1,vhit(2)) = MIN( sect%rect(1:3,1,vhit(1)), sect%rect(1:3,1,vhit(2)) )
        sect%rect(1:3,2,vhit(2)) = MAX( sect%rect(1:3,2,vhit(1)), sect%rect(1:3,2,vhit(2)) )
        sect%vol(vhit(2)) = PRODUCT( sect%rect(1:3,2,vhit(2)) - sect%rect(1:3,1,vhit(2)) + 1 )
        sect%nrect = sect%nrect - 1
        if(vhit(1)<sect%nrect+1) then
          ! move last rect to vhit(1)
          sect%rect(1:3,1:2,vhit(1)) = sect%rect(1:3,1:2,sect%nrect+1)
          sect%vol(vhit(1)) = sect%vol(sect%nrect+1)
          ! update uvol for the moved rect
          if(SECTION_RECT_MAX==4) then
            if(vhit(1)==2) then     ! 1 is the unified rect, 4 moved to 2, 3 is untouched
              sect%uvol(3,2) = sect%uvol(4,3)
            elseif(vhit(2)==2) then ! 2 is the unified rect, 4 moved to 3, 1 is untouched
              sect%uvol(3,1) = sect%uvol(4,1)
            else                    ! 1 is the unified rect, 4 moved to 3, 2 is untouched
              sect%uvol(3,2) = sect%uvol(4,2)
            end if
          else
            call section_add_update_uvol(sect, vhit(1))
          end if
        end if
        call section_add_update_uvol(sect, vhit(2))
        hit = vhit(2)
      else
        exit
      end if
    end do

  end subroutine section_add_check_unify


  !! Return .true. if intersection of $sect with cuboid $rect is not empty.
  !!
  !! Optionally, an offset used for the upper bound of $sect can be given.
  !!
  logical function section_intersect(sect, rect, opt_offset)
    type(section), intent(inout) :: sect         !! the section
    integer(i_k),  intent(in) :: rect(3,2)       !! 3D cuboid to intersect
    integer(i_k),  intent(in) :: opt_offset(3)   !! optional offset for upper bound of $sect
    optional :: opt_offset
    
    integer(i_k) :: intersec(3,2), irect

    section_intersect = .false.
    
    if(present(opt_offset)) then
    
      do irect = 1, sect%nrect
        ! intersect sect%rect with rect
        intersec(1:3,1) = MAX( sect%rect(1:3,1,irect)                , rect(1:3,1) )
        intersec(1:3,2) = MIN( sect%rect(1:3,2,irect)+opt_offset(1:3), rect(1:3,2) )
        
        !write(*,'(2(I3),A,3(I5),A,3(I5))') irect,sect%nrect,': ',sect%rect(1:3,1,irect), ' - ',sect%rect(1:3,2,irect)
        
        if( ALL( intersec(1:3,1) <= intersec(1:3,2) ) ) then
          ! intersection is not empty
          section_intersect = .true.
          return
        end if
      end do
    
    else

      do irect = 1, sect%nrect
        ! intersect sect%rect with rect
        intersec(1:3,1) = MAX( sect%rect(1:3,1,irect), rect(1:3,1) )
        intersec(1:3,2) = MIN( sect%rect(1:3,2,irect), rect(1:3,2) )
        if( ALL( intersec(1:3,1) <= intersec(1:3,2) ) ) then
          ! intersection is not empty
          section_intersect = .true.
          return
        end if
      end do
      
    end if
    
  end function section_intersect



  !! Return .true. if $sect contains the point at $pos.
  !!
  logical function section_contains(sect, pos)
    type(section), intent(inout) :: sect         !! the section
    integer(i_k),  intent(in) :: pos(3)          !! pos
    
    integer(i_k) :: irect

    section_contains = .false.
    
    do irect = 1, sect%nrect
      if( ALL(pos(1:3) >= sect%rect(1:3,1,irect)) .and. ALL(pos(1:3) <= sect%rect(1:3,2,irect)) ) then
        section_contains = .true.
        return
      end if
    end do
    
  end function section_contains


  !! Calculate total volume of the section.
  !!
  pure subroutine section_volume(sect, volume)
    type(section), intent(inout) :: sect   !! the section
    integer(i_k),  intent(out) :: volume   !! volume
    integer(i_k) :: irect
      
    volume = 0
    do irect = 1, sect%nrect
      volume = volume + PRODUCT( sect%rect(1:3,2,irect) - sect%rect(1:3,1,irect) + 1 )
    end do
  
  end subroutine section_volume


end module section_mod
